package jj2000.j2k.encryption;

import jj2000.j2k.wavelet.analysis.*;
import jj2000.j2k.entropy.encoder.*;
import jj2000.j2k.codestream.*;
import jj2000.j2k.codestream.reader.CBlkInfo;
import jj2000.j2k.codestream.reader.PktCblksInfo;
import jj2000.j2k.codestream.writer.BitOutputBuffer;
import jj2000.j2k.codestream.writer.TagTreeEncoder;
import jj2000.j2k.encoder.*;
import jj2000.j2k.wavelet.*;
import jj2000.j2k.image.*;
import jj2000.j2k.util.*;
import jj2000.j2k.*;

import java.util.Iterator;
import java.util.Random;
import java.util.Vector;

/**
 * This class builds packets and keeps the state information of packet
 * interdependencies. It also supports saving the state and reverting
 * (restoring) to the last saved state, with the save() and restore() methods.
 *
 * <P>Each time the encodePacket() method is called a new packet is encoded,
 * the packet header is returned by the method, and the packet body can be
 * obtained with the getLastBodyBuf() and getLastBodyLen() methods.
 * */
public class FakePktEncoder {
	
	private HeaderTransformer ht;
    
    private Random random;
    
    /** The prefix for packet encoding options: 'P' */
    public final static char OPT_PREFIX = 'P';

    /** The list of parameters that is accepted for packet encoding.*/
    private final static String [][] pinfo = {
        { "Psop", "[<tile idx>] on|off"+
         "[ [<tile idx>] on|off ...]",
          "Specifies whether start of packet (SOP) markers should be used. "+
          "'on' enables, 'off' disables it.","off"},
        { "Peph", "[<tile idx>] on|off"+
         "[ [<tile  idx>] on|off ...]",
          "Specifies whether end of packet header (EPH) markers should be "+
          " used. 'on' enables, 'off' disables it.","off"}
    };

    /** The initial value for the lblock */
    private final static int INIT_LBLOCK = 3;
    

    /**
     * The tag tree for inclusion information. The indexes are outlined
     * below. Note that the layer indexes start at 1, therefore, the layer
     * index minus 1 is used. The subband indices are used as they are defined
     * in the Subband class. The tile indices start at 0 and follow a
     * lexicographical order.
     *
     * <ul>
     * <li>1st index: tile index, in lexicographical order</li>
     * <li>2nd index: component index </li>
     * <li>3rd index: resolution level </li>
     * <li>4th index: subband index </li>
     * <li>5th index: precinct index </li>
     * </ul>
     **/
    public TagTreeEncoder ttIncl[][][][][];

    /**
     * The tag tree for the maximum significant bit-plane. The indexes are
     * outlined below. Note that the layer indexes start at 1, therefore, the
     * layer index minus 1 is used. The subband indices are used as they are
     * defined in the Subband class. The tile indices start at 0 and follow a
     * lexicographical order.
     *
     * <ul>
     * <li>1st index: tile index, in lexicographical order</li>
     * <li>2nd index: component index </li>
     * <li>3rd index: resolution level </li>
     * <li>4th index: subband index - subband index offset </li>
     * <li>5th index: precinct index </li>
     * </ul>
     * */
    public TagTreeEncoder ttMaxBP[][][][][];

    /**
     * The base number of bits for sending code-block length information
     * (referred as Lblock in the JPEG 2000 standard). The indexes are
     * outlined below. Note that the layer indexes start at 1, therefore, the
     * layer index minus 1 is used. The subband indices are used as they are
     * defined in the Subband class. The tile indices start at 0 and follow a
     * lexicographical order.
     *
     * <ul>
     * <li>1st index: tile index, in lexicographical order </li>
     * <li>2nd index: component index </li>
     * <li>3rd index: resolution level </li>
     * <li>4th index: subband index - subband index offset </li>
     * <li>5th index: code-block index, in lexicographical order</li>
     * </ul>
     * */
    private int lblock[][][][][];

    /**
     * The last encoded truncation point for each code-block. A negative value
     * means that no information has been included for the block, yet. The
     * indexes are outlined below. The subband indices are used as they are
     * defined in the Subband class. The tile indices start at 0 and follow a
     * lexicographical order. The code-block indices follow a lexicographical
     * order within the subband tile.
     *
     * <P>What is actually stored is the index of the element in
     * CBlkRateDistStats.truncIdxs that gives the real truncation point.
     *
     * <ul>
     * <li>1st index: tile index, in lexicographical order </li>
     * <li>2nd index: component index </li>
     * <li>3rd index: resolution level </li>
     * <li>4th index: subband index - subband index offset </li>
     * <li>5th index: code-block index, in lexicographical order </li>
     * </ul>
     *  */
    private int prevtIdxs[][][][][];

    /**
     * The saved base number of bits for sending code-block length
     * information. It is used for restoring previous saved state by
     * restore(). The indexes are outlined below. Note that the layer indexes
     * start at 1, therefore, the layer index minus 1 is used. The subband
     * indices are used as they are defined in the Subband class. The tile
     * indices start at 0 and follow a lexicographical order.
     * 
     * <ul> 
     * <li>1st index: tile index, in lexicographical order </li>
     * <li>2nd index: component index </li>
     * <li>3rd index: resolution level </li>
     * <li>4th index: subband index - subband index offset </li>
     * <li>5th index: code-block index, in lexicographical order</li>
     * </ul>
     * */
    private int bak_lblock[][][][][];

    /**
     * The saved last encoded truncation point for each code-block. It is used
     * for restoring previous saved state by restore(). A negative value means
     * that no information has been included for the block, yet. The indexes
     * are outlined below. The subband indices are used as they are defined in
     * the Subband class. The tile indices start at 0 and follow a
     * lexicographical order. The code-block indices follow a lexicographical
     * order within the subband tile.
     *
     * <ul>
     * <li>1st index: tile index, in lexicographical order </li>
     * <li>2nd index: component index </li>
     * <li>3rd index: resolution level </li>
     * <li>4th index: subband index - subband index offset </li>
     * <li>5th index: code-block index, in lexicographical order </li>
     * </ul>
     *  */
    private int bak_prevtIdxs[][][][][];

    /** The body buffer of the last encoded packet */
    private byte[] lbbuf;

    /** The body length of the last encoded packet */
    private int lblen;

    /** The saved state */
    private boolean saved;

    /** Whether or not there is ROI information in the last encoded Packet */
    private boolean roiInPkt = false;
    
    /** Length to read in current packet body to get all the ROI information */
    private int roiLen = 0;

    /**
     *  Array containing the coordinates, width, height, indexes, ... of the
     *  precincts
     *
     * <ul>
     * <li> 1st dim: tile index.</li>
     * <li> 2nd dim: component index.</li>
     * <li> 3nd dim: resolution level index.</li>
     * <li> 4th dim: subband index.</li>
     * <li> 5th dim: precinct index.</li>
     * </ul> 
     * */
    private PrecCoordInfo precArrayI[][][][][];
    
    /**
     *  Array containing the coordinates, width, height, indexes, ... of the
     *  code-blocks
     *
     * <ul>
     * <li> 1st dim: tile index.</li>
     * <li> 2nd dim: component index.</li>
     * <li> 3nd dim: resolution level index.</li>
     * <li> 4th dim: subband index.</li>
     * <li> 5th dim: code-block index.</li>
     * </ul> */
    private CBlkCoordInfo cbArrayI[][][][][];
    
    /** 
     * Maximum number of precincts in each resolution level :
     *
     * <ul>
     * <li> 1st dim: tile index.</li>
     * <li> 2nd dim: component index.</li>
     * <li> 3nd dim: resolution level index.</li>
     * </ul>
     */
    private Coord maxNumPrec[][][];
    
    
    /** 
     * The maximum number of resolution level for each component and each tile
     * */
    int mrl[][];    
    
    /** 
     * Keeps the first and the last subband index in the associated resolution
     * level for each tile and each component
     * */
    private int[][][][] subRange;     
    
    /** Whether or not the current packet is writable */
    private boolean packetWritable;
    
    

    public FakePktEncoder(int maxReslvl, HeaderTransformer ht) {
    	
    	this.ht = ht;    	
    	           
        Coord orTCoord;  // Original tile X and Y
        Coord tmpCoord = null;
        int numcb;       // Number of code-blocks
        Vector cblks = null;
        
        random = new Random (20011978l);

        
        // Get number of components and tiles
        int ncomp = 1;
        int ntiles = 1;

        // Do initial allocation
        ttIncl = new TagTreeEncoder[ntiles][ncomp][][][];
        ttMaxBP = new TagTreeEncoder[ntiles][ncomp][][][];
        lblock = new int[ntiles][ncomp][][][];
        prevtIdxs = new int[ntiles][ncomp][][][];

        cbArrayI = new CBlkCoordInfo[ntiles][ncomp][][][];
        precArrayI = new PrecCoordInfo[ntiles][ncomp][][][];
        mrl = new int[ntiles][ncomp];
        subRange = new int[ntiles][ncomp][][];


        // Finish allocation
        for(int t=0; t<ntiles; t++){ // Loop on tiles
            for(int c=0; c<ncomp; c++){ // Loop on components
                // Get number of resolution levels                
                mrl[t][c] = maxReslvl;
                
                subRange[t][c] = findSubInResLvl(c,t);
                

                lblock[t][c] = new int[mrl[t][c]+1][][];
                ttIncl[t][c] = new TagTreeEncoder[mrl[t][c]+1][][];
                ttMaxBP[t][c] = new TagTreeEncoder[mrl[t][c]+1][][];
                prevtIdxs[t][c] = new int[mrl[t][c]+1][][];
                
                cbArrayI[t][c] = new CBlkCoordInfo[mrl[t][c]+1][][];
                precArrayI[t][c] = new PrecCoordInfo[mrl[t][c]+1][][];
                
                for(int r=mrl[t][c]; r>=0; r--){ // Loop on resolution levels
                    
                    int maxPrec = 1; 
                      
                
                    // Allocate for subbands
                    ttIncl[t][c][r] = 
                        new TagTreeEncoder[subRange[t][c][r][1]+1][maxPrec];
                    ttMaxBP[t][c][r] = 
                        new TagTreeEncoder[subRange[t][c][r][1]+1][maxPrec];
                    prevtIdxs[t][c][r] = new int[subRange[t][c][r][1]+1][];
                    lblock[t][c][r] = new int[subRange[t][c][r][1]+1][];
                    cbArrayI[t][c][r] = 
                        new CBlkCoordInfo[subRange[t][c][r][1]+1][];
                    precArrayI[t][c][r] = 
                        new PrecCoordInfo[subRange[t][c][r][1]+1][];
                    

                    
                    for(int subIdx=subRange[t][c][r][0];
                        subIdx <= subRange[t][c][r][1]; subIdx ++){
                            // Loop on subbands

                        //numcb = tmpCoord.x*tmpCoord.y;
                    	numcb = ht.getDataCollector()
                    		.getPacketsInResolution(r).get(0)
                    		.getCblksInSubband(subIdx).size();

                    
                        lblock[t][c][r][subIdx] = new int[numcb];
                        ArrayUtil.intArraySet(lblock[t][c][r][subIdx],
                                              INIT_LBLOCK);

                        prevtIdxs[t][c][r][subIdx] = new int[numcb];
                        ArrayUtil.intArraySet(prevtIdxs[t][c][r][subIdx],-1);
                        
                        // For each precinct, get the number of code-blocks in
                        // the packet and create the tag trees
                        for(int pIdx=0 ; pIdx<maxPrec ; pIdx++){
                            
                            cblks = getCBlkInPrec(t,c,r,subIdx,pIdx,cblks);
                            
                            // the first element contains the number of
                            // code-blocks in the precinct in each direction
                            tmpCoord = (Coord)cblks.elementAt(0);
                            
                            ttIncl[t][c][r][subIdx][pIdx] =
                                new TagTreeEncoder(tmpCoord.y,tmpCoord.x);
                            (ttIncl[t][c][r][subIdx][pIdx]).sbIdx = subIdx;
                            (ttIncl[t][c][r][subIdx][pIdx]).resLvl = r;
                            
                            ttMaxBP[t][c][r][subIdx][pIdx] =
                                new TagTreeEncoder(tmpCoord.y,tmpCoord.x);
                        }
                    }
                }
            }
        }
    }
    
    
    /**
     * Returns the code-blocks contained in the precinct which index is
     * precIdx for the tile 't', the component 'c', the resolution level 'r'
     * and the subband index 'sbIdx'. The result is returned in a Coord object
     * which is allocated if it is null. The returned object is a Vector
     * object whose first element is a Coord object containing the number of
     * code-blocks in each direction and then, cblks contains all the
     * code-blocks indexes.
     *
     * @param t the tile index
     *
     * @param c the component index
     *
     * @param r the resolution level
     *
     * @param sbIdx the subband index
     * 
     * @param precIdx the precinct index
     *
     * @param cblks the Vector object in which to return the code-blocks
     * indexes
     *
     * @return the code-blocks contained in the precinct as the first Coord
     * object stored in the Vector and then the code-blocks indexes
     * */
    public Vector getCBlkInPrec(int t, int c, int r, int sbIdx, 
            int precIdx, Vector cblks) {
            
        int pulx, puly, pw, ph;
        int cbIdx, culx, culy, cw, ch;
        int maxulx, maxuly;
        Coord numCB = new Coord();
        CBlkCoordInfo cbI;
        PrecCoordInfo pktI;
        //SubbandAn curSub = null;
        
        //curSub = (SubbandAn)savedSbTree[t][c].getSubbandByIdx(r,sbIdx);
        
        // Allocated cblks if it is null
        if ( cblks==null ) {
            cblks = new Vector();
        }
        else {
            // If reused, empty Vector object
            cblks.removeAllElements();
        }
        
        //packets in r and s
   		
   		Vector<CBlkInfo> relevantCodeblocks = new Vector<CBlkInfo>();
   		
   		PktCblksInfo pci  = ht.getDataCollector().getPacketsInResolution(r).get(0);
   		
   		int i = 0;
   		Iterator <CBlkInfo> it = pci.getAllCBlks().iterator();
   		while (it.hasNext()) {   			
   			CBlkInfo ccb = it.next();
   			if (ccb.sbIdx == sbIdx) {
   				relevantCodeblocks.add(ccb);
   				cblks.add(new Integer(i));
   				i++;
   			}
   		}   		
        
        if( relevantCodeblocks.size() == 0 ){
            cblks.addElement(new Coord(0,0));
            return cblks;
        }        
        
        // only square blocks!!!
        numCB.x = numCB.y = (int) (Math.sqrt(cblks.size()));
        
        // Insert the number of code-blocks contained in a Coord object at the
        // begining of the vector
        cblks.insertElementAt(numCB, 0);
        return cblks;
    }

    /**
     * Encodes a packet and returns the buffer containing the encoded packet
     * header. The code-blocks appear in a 3D array of CBlkRateDistStats,
     * 'cbs'. The first index is the tile index in lexicographical order, the
     * second index is the subband index (as defined in the Subband class),
     * and the third index is the code-block index (whithin the subband tile)
     * in lexicographical order as well. The indexes of the new truncation
     * points for each code-block are specified by the 3D array of int
     * 'tIndx'. The indices of this array are the same as for cbs. The
     * truncation point indices in 'tIndx' are the indices of the elements of
     * the 'truncIdxs' array, of the CBlkRateDistStats class, that give the
     * real truncation points. If a truncation point index is negative it
     * means that the code-block has not been included in any layer yet. If
     * the truncation point is less than or equal to the highest truncation
     * point used in previous layers then the code-block is not included in
     * the packet. Otherwise, if larger, the code-block is included in the
     * packet. The body of the packet can be obtained with the
     * getLastBodyBuf() and getLastBodyLen() methods.
     *
     * <P>Layers must be coded in increasing order, in consecutive manner, for
     * each tile, component and resolution level (e.g., layer 1, then layer 2,
     * etc.). For different tile, component and/or resolution level no
     * particular order must be followed.
     *
     * @param ly The layer index (starts at 1).
     *
     * @param c The component index.
     *
     * @param r The resolution level
     *
     * @param t Index of the current tile
     *
     * @param cbs The 3D array of coded code-blocks.
     *
     * @param tIndx The truncation point indices for each code-block.
     *
     * @param hbuf The header buffer. If null a new BitOutputBuffer is created
     * and returned. This buffer is reset before anything is written to it.
     *
     * @param bbuf The body buffer. If null a new one is created. If not large
     * enough a new one is created.
     *
     * @param pIdx The precinct index
     *
     * @return The buffer containing the packet header.
     * */
    public BitOutputBuffer encodePacket(PktCblksInfo pci, BitOutputBuffer hbuf) {
    	
    	
    	int c = 0;
    	int t = 0;
    	int pIdx = 0; //precinct idx!
    	
    	int r = pci.getResolution();
    	int ly = pci.getLayer()+1;

    	
    	int b,by,bx,s,i,maxi;
    	int ncb,ncbx,ncby;
    	int thmax;
    	int newtp;
    	int cblen;
    	int prednbits,nbits,deltabits;
    	TagTreeEncoder cur_ttIncl,cur_ttMaxBP;
    	int cur_prevtIdxs[]; // last encoded truncation points
    	CBlkRateDistStats cur_cbs[];
    	int cur_tIndx[]; // truncation points to encode
    	int minsb,maxsb;
    	Vector cbv = null;
    	Coord cbCoord = null;
    	
    	Vector<CBlkInfo> cblks = pci.getAllCBlks();
    	
    	int firstCB_horIdx  = -1;
    	int firstCB_vertIdx = -1;
    	boolean precIdxExist = false;
    	roiInPkt = false;
    	roiLen = 0;

    	maxsb = subRange[t][c][r][1]+1;
    	minsb = subRange[t][c][r][0];
    	
    	if(hbuf == null){
    		hbuf = new BitOutputBuffer();
    	}
    	else{
    		hbuf.reset();
    	}

    	// Invalidate last body buffer
    	lbbuf = null;
    	lblen = pci.getPacketBodyLength();

    	// Signal that packet is present    	
    	hbuf.writeBit(1);

    	
    	for(s=minsb; s<maxsb; s++){ // Loop on subbands

    		// If we arrive here, it means that the precinct exists in this
    		// subband.
    		precIdxExist = true;

    		// Get in the vector cbv the code-blocks that "belong" to the
    		// precinct whose index is 'pIdx'
    		cbv = getCBlkInPrec(t,c,r,s,pIdx,cbv);

    		cur_ttIncl = ttIncl[t][c][r][s][pIdx];
    		cur_ttMaxBP = ttMaxBP[t][c][r][s][pIdx];
    		cur_prevtIdxs = prevtIdxs[t][c][r][s];
    		ncbx = cur_ttIncl.getWidth();
    		ncby = cur_ttIncl.getHeight();

    		// Set tag tree values for this subband 
    		Iterator<CBlkInfo> it = pci.getCblksInSubband(s).iterator();
    		int numcb = 0;
    		//for( int numcb=1 ; numcb<cbv.size() ; numcb++ ){
    		while (it.hasNext()) {
    			CBlkInfo ccb = it.next();
    			numcb ++;
    			//b = ((Integer)cbv.elementAt(numcb)).intValue();
    			//cbCoord = cbArrayI[t][c][r][s][b].idx;
    			
    			bx = ccb.coord.x;
    			by = ccb.coord.y;

    			// Save "offset" as tag trees are indexed from 0
    			if ( numcb==1 ) {
    				firstCB_horIdx = bx;
    				firstCB_vertIdx = by;
    			}

    			if (ccb.inclusionInfo[ly-1] == CBlkInfo.InclusionInfo.FI) { // First inclusion
    				// Set layer index
    				cur_ttIncl.setValue(by-firstCB_vertIdx,
    						bx-firstCB_horIdx, ly-1); 
    			}
    		}
    		
    		

    		if (ly == 1) { // First layer, need to set the skip of MSBP
        		it = pci.getCblksInSubband(s).iterator();
    			//for ( int numcb=1 ; numcb<cbv.size() ; numcb++ ) {
        		while (it.hasNext()) {
        			CBlkInfo ccb = it.next();
    				//b = ((Integer)cbv.elementAt(numcb)).intValue();
    				//cbCoord = cbArrayI[t][c][r][s][b].idx;
    				bx = ccb.coord.x;
    				by = ccb.coord.y;

    				// Save "offset" as tag trees are indexed from 0
    				if ( numcb==1 ) {
    					firstCB_horIdx = bx;
    					firstCB_vertIdx = by;
    				}
 
    				// Set skip of most sig. bit-plane
    				cur_ttMaxBP.setValue(by-firstCB_vertIdx,
    						bx-firstCB_horIdx,
    						ccb.msbSkipped
    						);
    			}
    		}  

    		// Now encode the information
    		it = pci.getCblksInSubband(s).iterator();

    		//for(int numcb=1; numcb<cbv.size(); numcb++){
    		b = -1;
    		while (it.hasNext()) {
    			//b = ((Integer)cbv.elementAt(numcb)).intValue();
    			//cbCoord = cbArrayI[t][c][r][s][b].idx;
    			//bx = cbCoord.x;
    			//by = cbCoord.y;
    			CBlkInfo ccb = it.next();
    			b++;
    		
    			bx = ccb.coord.x;
    			by = ccb.coord.y;


    			// Save "offset" as tag trees are indexed from 0
    			if( numcb==1 ){
    				firstCB_horIdx = bx;
    				firstCB_vertIdx = by;
    			}

    			// 1) Inclusion information
    			if (ccb.inclusionInfo[ly-1] == CBlkInfo.InclusionInfo.FI ||
    					ccb.inclusionInfo[ly-1] == CBlkInfo.InclusionInfo.PI) {
    				// Block included in this precinct
    				if (ccb.inclusionInfo[ly-1] == CBlkInfo.InclusionInfo.FI ) { // First inclusion
    					// Encode layer info
    					cur_ttIncl.encode(by-firstCB_vertIdx,
    							bx-firstCB_horIdx,ly,hbuf);

    					// 2) Max bitdepth info. Encode value
    					
    					int val = ccb.msbSkipped;
    					thmax = val+1;    					
    					for (i=1; i<=thmax;i++) {
    						cur_ttMaxBP.encode(by-firstCB_vertIdx,
    								bx-firstCB_horIdx,i,hbuf);
    					}

    					// Count body size for packet
    					//lblen += cur_cbs[b].
    					//truncRates[cur_cbs[b].truncIdxs[cur_tIndx[b]]];
    				}
    				else { // Already in previous layer
    					// Send "1" bit
    					hbuf.writeBit(1);
    					// Count body size for packet
    					/*lblen +=
    						cur_cbs[b].
    						truncRates[cur_cbs[b].truncIdxs[cur_tIndx[b]]] -
    						cur_cbs[b].
    						truncRates[cur_cbs[b].
    						           truncIdxs[cur_prevtIdxs[b]]];*/    					
    				}

    				// 3) Truncation point information
    				/*if (cur_prevtIdxs[b]<0) {
    					newtp = cur_cbs[b].truncIdxs[cur_tIndx[b]];
    				}
    				else {
    					newtp = cur_cbs[b].truncIdxs[cur_tIndx[b]]-
    					cur_cbs[b].truncIdxs[cur_prevtIdxs[b]]-1;
    				}*/
    				newtp = ccb.ntp[ly-1]-1; //neeed this
    				
    				if (newtp < 0 ) {System.exit (1); };

    				// Mix of switch and if is faster
    				switch (newtp) {
    				case 0:
    					hbuf.writeBit(0); // Send one "0" bit
    					break;
    				case 1:
    					hbuf.writeBits(2,2); // Send one "1" and one "0"
    					break;
    				case 2:
    				case 3:
    				case 4:
    					// Send two "1" bits followed by 2 bits
    					// representation of newtp-2
    					hbuf.writeBits((3<<2)|(newtp-2),4);
    					break;
    				default:
    					if (newtp <= 35) {
    						// Send four "1" bits followed by a five bits
    						// representation of newtp-5
    						hbuf.writeBits((15<<5)|(newtp-5),9);
    					}
    					else if (newtp <= 163) {
    						// Send nine "1" bits followed by a seven bits
    						// representation of newtp-36
    						hbuf.writeBits((511<<7)|(newtp-36),16);
    					}
    					else {
    						throw new
    						ArithmeticException("Maximum number "+
    								"of truncation "+
    						"points exceeded");
    					}
    				}
    			}
    			else { // Block not included in this layer
    				if (ccb.inclusionInfo[ly-1] == CBlkInfo.InclusionInfo.PN) {
    					// Already in previous layer. Send "0" bit
    					hbuf.writeBit(0);
    				}
    				else { // Not in any previous layers
    					cur_ttIncl.encode(by-firstCB_vertIdx,
    							bx-firstCB_horIdx,ly,hbuf);
    				}
    				// Go to the next one.
    				continue;
    			}

    			// Code-block length

    			// We need to compute the maximum number of bits needed to
    			// signal the length of each terminated segment and the final
    			// truncation point.
    			/*newtp = 1;
    			maxi = cur_cbs[b].truncIdxs[cur_tIndx[b]];
    			cblen = (cur_prevtIdxs[b]<0) ? 0 :
    				cur_cbs[b].truncRates[cur_cbs[b].
    				                      truncIdxs[cur_prevtIdxs[b]]];

    			// Loop on truncation points
    			i = (cur_prevtIdxs[b]<0) ? 0 :
    				cur_cbs[b].truncIdxs[cur_prevtIdxs[b]]+1;
    			int minbits = 0;
    			for (; i<maxi; i++, newtp++) {
    				// If terminated truncation point calculate length
    				if (cur_cbs[b].isTermPass != null &&
    						cur_cbs[b].isTermPass[i]) {

    					// Calculate length
    					cblen = cur_cbs[b].truncRates[i] - cblen;

    					// Calculate number of needed bits
    					prednbits = lblock[t][c][r][s][b] +
    					MathUtil.log2(newtp);
    					minbits = ((cblen>0) ? MathUtil.log2(cblen) : 0)+1;

    					// Update Lblock increment if needed
    					for(int j=prednbits; j<minbits; j++) {
    						lblock[t][c][r][s][b]++;
    						hbuf.writeBit(1);
    					}
    					// Initialize for next length
    					newtp = 0;
    					cblen = cur_cbs[b].truncRates[i];
    				}
    			}*/
    			newtp = ccb.ntp[ly-1]; // now I need "the real" ntp

    			// Last truncation point length always sent

    			// Calculate length
    			//cblen = cur_cbs[b].truncRates[i] - cblen;
    			cblen = ccb.len[ly-1];

    			// Calculate number of bits
    			
    			
    			//System.out.println ("lblock[t][c][r][s][b]" + lblock[t][c][r][s][b]);
    			
    			prednbits = lblock[t][c][r][s][b] + MathUtil.log2(newtp);
    			int minbits = ((cblen>0) ? MathUtil.log2(cblen) : 0)+1;
    			// Update Lblock increment if needed
    			for(int j=prednbits; j<minbits; j++) {
    				lblock[t][c][r][s][b]++;
    				hbuf.writeBit(1);
    			}
    			//System.out.println ("lblock[t][c][r][s][b]" + lblock[t][c][r][s][b]);

    			// End of comma-code increment
    			hbuf.writeBit(0);

    			
    			/*// There can be terminated several segments, send length info
    			// for all terminated truncation points in addition to final
    			// one
    			
    			//BUT NOT FOR US HERE :-)
    			newtp = 1;
    			maxi = cur_cbs[b].truncIdxs[cur_tIndx[b]];
    			cblen = (cur_prevtIdxs[b]<0) ? 0 :
    				cur_cbs[b].truncRates[cur_cbs[b].
    				                      truncIdxs[cur_prevtIdxs[b]]];
    			// Loop on truncation points and count the groups
    			i = (cur_prevtIdxs[b]<0) ? 0 :
    				cur_cbs[b].truncIdxs[cur_prevtIdxs[b]]+1;
    			for (; i<maxi; i++, newtp++) {
    				// If terminated truncation point, send length
    				if (cur_cbs[b].isTermPass != null &&
    						cur_cbs[b].isTermPass[i]) {

    					cblen = cur_cbs[b].truncRates[i] - cblen;
    					nbits = MathUtil.log2(newtp)+lblock[t][c][r][s][b];
    					hbuf.writeBits(cblen,nbits);

    					// Initialize for next length
    					newtp = 0;
    					cblen = cur_cbs[b].truncRates[i];
    				}
    			}*/
    			
    			// Last truncation point length is always signalled
    			// First calculate number of bits needed to signal
    			// Calculate length
    			//cblen = cur_cbs[b].truncRates[i] - cblen;
    			
    			nbits = MathUtil.log2(newtp) + lblock[t][c][r][s][b];
    			hbuf.writeBits(cblen,nbits);
    			//b++; //darf da nicht stehen wegen continue f***
    			
    			//System.out.println ("ccb " + ccb.toString());
    			//System.out.println ("written " + hbuf.toBitString());
    			
    			

    		} // End loop on code-blocks
    	} // End loop on subband

    	// -> Copy the data to the body buffer
    	//NO WE DONT

    	if(!precIdxExist){ // No precinct found in this resolution level
    		packetWritable = false;
    		return hbuf;
    	} else {
    		packetWritable = true;
    	}

    	// Must never happen
    	if(hbuf.getLength()==0){
    		throw new Error("You have found a bug in PktEncoder, method:"+
    		" encodePacket");
    	}

    	return hbuf;
    }

    /**
     * Returns the buffer of the body of the last encoded packet. The length
     * of the body can be retrieved with the getLastBodyLen() method. The
     * length of the array returned by this method may be larger than the
     * actual body length.
     *
     * @return The buffer of body of the last encoded packet.
     *
     * @exception IllegalArgumentException If no packet has been coded since
     * last reset(), last restore(), or object creation.
     *
     * @see #getLastBodyLen
     * */
    public byte[] getLastBodyBuf() {
        if (lbbuf == null) {
            throw new IllegalArgumentException();
        }
        return lbbuf;
    }

    /**
     * Returns the length of the body of the last encoded packet, in
     * bytes. The body itself can be retrieved with the getLastBodyBuf()
     * method.
     *
     * @return The length of the body of last encoded packet, in bytes.
     *
     * @see #getLastBodyBuf
     * */
    public int getLastBodyLen() {
        return lblen;
    }

    /**
     * Saves the current state of this object. The last saved state
     * can be restored with the restore() method.
     *
     * @see #restore
     * */
    public void save() {
        int s,r,n,t,p;
        int maxsbi,minsbi;

        // Have we done any save yet?
        if (bak_lblock == null) {
            // Allocate backup buffers
            bak_lblock = new int[ttIncl.length][][][][];
            bak_prevtIdxs = new int[ttIncl.length][][][][];
            for (t=ttIncl.length-1; t>=0; t--) {
                bak_lblock[t] = new int[ttIncl[t].length][][][];
                bak_prevtIdxs[t] = new int[ttIncl[t].length][][][];
                for (n=ttIncl[t].length-1; n>=0; n--) {
                    bak_lblock[t][n] = new int[lblock[t][n].length][][];
                    bak_prevtIdxs[t][n] = new int[ttIncl[t][n].length][][];
                    for (r=lblock[t][n].length-1; r>=0; r--) {
                        bak_lblock[t][n][r] =
                            new int[lblock[t][n][r].length][];
                        bak_prevtIdxs[t][n][r] =
                            new int[ttIncl[t][n][r].length][];
                        maxsbi = ttIncl[t][n][r].length;
                        minsbi = maxsbi>>2; // minsbi = maxsbi/4;
                        for (s = minsbi; s<maxsbi; s++) {
                            bak_lblock[t][n][r][s] = 
                                new int[lblock[t][n][r][s].length];
                            bak_prevtIdxs[t][n][r][s] =
                                new int[prevtIdxs[t][n][r][s].length];
                        }
                    }
                }
            }
        }

        //-- Save the data

        // Use reference caches to minimize array access overhead
        TagTreeEncoder
            ttIncl_t_n[][][],
            ttMaxBP_t_n[][][],
            ttIncl_t_n_r[][],
            ttMaxBP_t_n_r[][];
        int
            lblock_t_n[][][],
            bak_lblock_t_n[][][],
            prevtIdxs_t_n_r[][],
            bak_prevtIdxs_t_n_r[][];
            
        // Loop on tiles
        for (t=ttIncl.length-1; t>=0; t--) {
            // Loop on components
            for (n=ttIncl[t].length-1; n>=0; n--) {
                // Initialize reference caches
                lblock_t_n = lblock[t][n];
                bak_lblock_t_n = bak_lblock[t][n];
                ttIncl_t_n = ttIncl[t][n];
                ttMaxBP_t_n = ttMaxBP[t][n];
                // Loop on resolution levels
                for (r=lblock_t_n.length-1; r>=0; r--) {
                    // Initialize reference caches
                    ttIncl_t_n_r = ttIncl_t_n[r];
                    ttMaxBP_t_n_r = ttMaxBP_t_n[r];
                    prevtIdxs_t_n_r = prevtIdxs[t][n][r];
                    bak_prevtIdxs_t_n_r = bak_prevtIdxs[t][n][r];

                    // Loop on subbands
                    maxsbi = ttIncl_t_n_r.length;
                    minsbi = maxsbi>>2; // minsbi = maxsbi/4;
                    for (s = minsbi; s<maxsbi; s++) {
                        // Save 'lblock'
                        System.arraycopy(lblock_t_n[r][s],0,
                                         bak_lblock_t_n[r][s],0,
                                         lblock_t_n[r][s].length);
                        // Save 'prevtIdxs'
                        System.arraycopy(prevtIdxs_t_n_r[s],0,
                                         bak_prevtIdxs_t_n_r[s],0,
                                         prevtIdxs_t_n_r[s].length);
                        if ( precArrayI[t][n][r][s] != null ) {
                            for (p=precArrayI[t][n][r][s].length-1 ; p>=0 ; p--)
                            {
                                if ( p<ttIncl_t_n_r[s].length ) {
                                    // Save tag trees
                                    ttIncl_t_n_r[s][p].save();
                                    ttMaxBP_t_n_r[s][p].save();
                                }
                            }
                        }
                    }
                }
            }
        }

        // Set the saved state
        saved = true;
    }

    /**
     * Restores the last saved state of this object. An
     * IllegalArgumentException is thrown if no state has been saved.
     *
     * @see #save
     * */
    public void restore() {
        int s,r,n,t,p;
        int maxsbi,minsbi;


        if (!saved) {
            throw new IllegalArgumentException();
        }

        // Invalidate last encoded body buffer
        lbbuf = null;

        //-- Restore tha data

        // Use reference caches to minimize array access overhead
        TagTreeEncoder
            ttIncl_t_n[][][],
            ttMaxBP_t_n[][][],
            ttIncl_t_n_r[][],
            ttMaxBP_t_n_r[][];
        int
            lblock_t_n[][][],
            bak_lblock_t_n[][][],
            prevtIdxs_t_n_r[][],
            bak_prevtIdxs_t_n_r[][];
            
        // Loop on tiles
        for (t=ttIncl.length-1; t>=0; t--) {
            // Loop on components
            for (n=ttIncl[t].length-1; n>=0; n--) {
                // Initialize reference caches
                lblock_t_n = lblock[t][n];
                bak_lblock_t_n = bak_lblock[t][n];
                ttIncl_t_n = ttIncl[t][n];
                ttMaxBP_t_n = ttMaxBP[t][n];
                // Loop on resolution levels
                for (r=lblock_t_n.length-1; r>=0; r--) {
                    // Initialize reference caches
                    ttIncl_t_n_r = ttIncl_t_n[r];
                    ttMaxBP_t_n_r = ttMaxBP_t_n[r];
                    prevtIdxs_t_n_r = prevtIdxs[t][n][r];
                    bak_prevtIdxs_t_n_r = bak_prevtIdxs[t][n][r];

                    // Loop on subbands
                    maxsbi = ttIncl_t_n_r.length;
                    minsbi = maxsbi>>2; // minsbi = maxsbi/4;
                    for (s = minsbi; s<maxsbi; s++) {
                        // Restore 'lblock'
                        System.arraycopy(bak_lblock_t_n[r][s],0,
                                         lblock_t_n[r][s],0,
                                         lblock_t_n[r][s].length);
                        // Restore 'prevtIdxs'
                        System.arraycopy(bak_prevtIdxs_t_n_r[s],0,
                                         prevtIdxs_t_n_r[s],0,
                                         prevtIdxs_t_n_r[s].length);
                        // Loop on precincts
                        if ( precArrayI[t][n][r][s] != null ) {
                            for(p=precArrayI[t][n][r][s].length-1 ; p>=0 ; p--) 
                            {
                                if ( p<ttIncl_t_n_r[s].length ) {
                                    // Restore tag trees
                                    ttIncl_t_n_r[s][p].restore();
                                    ttMaxBP_t_n_r[s][p].restore();
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Resets the state of the object to the initial state, as if the object
     * was just created.
     * */
    public void reset() {
        int s,r,n,t,p;
        int maxsbi,minsbi;

        // Invalidate save
        saved = false;
        // Invalidate last encoded body buffer
        lbbuf = null;

        // Reinitialize each element in the arrays

        // Use reference caches to minimize array access overhead
        TagTreeEncoder
            ttIncl_t_n[][][],
            ttMaxBP_t_n[][][],
            ttIncl_t_n_r[][],
            ttMaxBP_t_n_r[][];
        int
            lblock_t_n[][][],
            prevtIdxs_t_n_r[][];
            
        // Loop on tiles
        for (t=ttIncl.length-1; t>=0; t--) {
            // Loop on components
            for (n=ttIncl[t].length-1; n>=0; n--) {
                // Initialize reference caches
                lblock_t_n = lblock[t][n];
                ttIncl_t_n = ttIncl[t][n];
                ttMaxBP_t_n = ttMaxBP[t][n];
                // Loop on resolution levels
                for (r=lblock_t_n.length-1; r>=0; r--) {
                    // Initialize reference caches
                    ttIncl_t_n_r = ttIncl_t_n[r];
                    ttMaxBP_t_n_r = ttMaxBP_t_n[r];
                    prevtIdxs_t_n_r = prevtIdxs[t][n][r];

                    // Loop on subbands
                    maxsbi = ttIncl_t_n_r.length;
                    minsbi = maxsbi>>2; // minsbi = maxsbi/4;
                    for (s = minsbi; s<maxsbi; s++) {
                        // Reset 'prevtIdxs'
                        ArrayUtil.intArraySet(prevtIdxs_t_n_r[s],-1);
                        
                        // Reset 'lblock'
                        ArrayUtil.intArraySet(lblock_t_n[r][s],
                                              INIT_LBLOCK);
                        // Loop on precincts
                        if ( precArrayI[t][n][r][s] != null ) {
                            for (p=precArrayI[t][n][r][s].length-1;p>=0;p--) {
                                if ( p<ttIncl_t_n_r[s].length ) {
                                    // Reset tag trees
                                    ttIncl_t_n_r[s][p].reset();
                                    ttMaxBP_t_n_r[s][p].reset();
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    

    
    /**
     * Finds the number of subbands in each resolution level according to the
     * decomposition tree for the specified tile and component. Since JPEG
     * 2000 part I only supports dyadic decomposition, there is one subbands
     * in resolution level 0 and 3 in each other (numbered from 1 to 3).
     *
     * @param c The component
     *
     * @param t the tile
     *
     * @return Minimun and maximum subband identifier for each resolution
     * level (First index= resolution level, second index = 0(minimum) or
     * 1(maximum))
     * */
    private int[][] findSubInResLvl(int c, int t){
        int[][] subRange = new int[mrl[t][c]+1][2];
        
        // Dyadic decomposition
        for(int i=mrl[t][c]; i>0; i--){
            subRange[i][0] = 1;
            subRange[i][1] = 3;
        }
        subRange[0][0] = 0;
        subRange[0][1] = 0;

        return subRange;
    }
    

    

    
  
    
    /** 
     * Returns the maximum resolution level for all tile-components. This
     * method is used by the rate allocator.
     * */
    public int[][] getMRL(){
        return mrl;
    }
    
    /** 
     * Returns true if the current packet is writable i.e. should be written.
     * Returns false otherwise.
     * */
    public boolean isPacketWritable() {
        return packetWritable;
    }

    /** 
     * Tells if there was ROI information in the last written packet 
     * */
    public boolean isROIinPkt(){
        return roiInPkt;
    }

    /** Gives the length to read in current packet body to get all ROI
     * information */
    public int getROILen(){
        return roiLen;
    }

    /**
     * Returns the parameters that are used in this class and implementing
     * classes. It returns a 2D String array. Each of the 1D arrays is for a
     * different option, and they have 3 elements. The first element is the
     * option name, the second one is the synopsis, the third one is a long
     * description of what the parameter is and the fourth is its default
     * value. The synopsis or description may be 'null', in which case it is
     * assumed that there is no synopsis or description of the option,
     * respectively. Null may be returned if no options are supported.
     *
     * @return the options name, their synopsis and their explanation, 
     * or null if no options are supported.
     * */
    public static String[][] getParameterInfo() {
        return pinfo;
    }
}
