/*
 * Created on Mar 13, 2007
 *
 */

package jj2000.j2k.encryption;

import java.io.ByteArrayInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collections;
import java.util.Iterator;
import java.util.Random;
import java.util.Vector;

import jj2000.j2k.codestream.Markers;
import jj2000.j2k.codestream.reader.CBlkInfo;
import jj2000.j2k.codestream.reader.DataCollector;
import jj2000.j2k.codestream.reader.HeaderDecoder;
import jj2000.j2k.codestream.reader.PktCblksInfo;
import jj2000.j2k.codestream.reader.PktHeaderBitReader;
import jj2000.j2k.codestream.reader.CBlkInfo.InclusionInfo;
import jj2000.j2k.codestream.writer.BitOutputBuffer;
import jj2000.j2k.codestream.writer.FileCodestreamWriter;
import jj2000.j2k.io.RandomAccessIO;
import jj2000.j2k.util.ArrayUtil;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;
import jj2000.j2k.util.ParameterList;


/**
 * this class collects information while a bitstream is read by PktDecoder. 
 * It can then be used to either transform the packet headers of this bitstream
 * and write a new header-encrypted bistream, or it can be used to decrypt a previously
 * encrypted bitstream (when <tt>decode</tt> is set to true).  
 * 
 * A better name for this class would have been somethin along the lines of HeaderTransformer,
 * as the inclusion information is only one part of the header information that can 
 * be transformed by this class.
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 *
 */
public class HeaderTransformer {
	
	public static final int MAXBP = 12;
	
	/** the associated data collector which holds the packet and codeblock info */
	private DataCollector dc;
	
	/** the random generator for the transformation passes */
	private Random random;
	
	/** the metaRandom generator that seeds the random generators for the transformation passes*/
	private Random metaRandom;
	
	/** the seed for <tt>metaRandom</tt>*/
	private long seed;

	/** maximum resolution level */
	private int maxRL;
	
	/** the resolution level at which  to start messing */
	private int startRL;

	/** Mathematica output */
	public String mathematica;

	/** deoding mode ? */
	private boolean decode = false;
	
	/** Set the amount of output produced by inclusion messer */
	public boolean verbose = false;
	
	/** the parameter list */
	private ParameterList pl;
	
	/** which headers to encrypt/decrypt */
	private String headers;
	
	/** the main Header (from the input stream, can be modified)*/
	private byte[] mainHeader;
	
	/** the data in the input stream after the last byte of the last
	 * packet. (usually just EOC) */
	private byte[] epilog;

	/** Reference to the stream where to read from */
	private RandomAccessIO ehs;
	
	/** use sop? */
	private boolean sop;
	
	/** use eph? */
	private boolean eph;
		
	private int startIdx;
	
	private int encBodyStopPIdx;
	
	

	/**
	 * create an inclusion messer. this can be used to encrypt header data and write a new 
	 * file or to decrypt a previously encrypted file
	 * 
	 * @param ehs the input stream 
	 * @param maxRL the maximum number of resolution levels
	 * @param seed seed for the random number generator
	 * @param pl the parameter list
	 */
	public HeaderTransformer(DataCollector dc, RandomAccessIO ehs, int maxRL, long seed, ParameterList pl) {
		//allCblks = new Vector<CBlkInfo>();
		this.dc = dc;
		this.ehs = ehs;		
		this.maxRL = maxRL;
		this.seed = seed;
		this.metaRandom = new Random(seed);
		this.maxRL = maxRL;
		this.mathematica = "";
		this.pl = pl;
		this.verbose = pl.getBooleanParameter("epic");
		this.headers = pl.getParameter("headers");
		this.startRL = pl.getIntParameter("startRes");
		this.sop = pl.getBooleanParameter("Psop");
		this.eph = pl.getBooleanParameter("Peph");		
		this.startIdx = pl.getIntParameter("startPIdx");
		this.encBodyStopPIdx = pl.getIntParameter("encBodyStopPIdx");
	}

	/**
	 * reset the random generator
	 *
	 */
	public void resetRandom() {
		this.random = new Random(seed);
	}
	
	/**
	 * read the main headers (main, tile and tile-part)
	 * from the input stream and save them to <tt>mainHeader</tt>.
	 * read the epilog (all data after the last byte of the last
	 * packet) from the stream and save it to <tt>epilog</tt>
	 *
	 */
	public void readMainHeadersAndEpilog () {
		if (verbose) System.out.println ("Reading Main Headers and Epilog");
		Vector<PktCblksInfo> pktCblkInfos = dc.getPktCblkInfos();  
		try {
			int hlength = pktCblkInfos.get(0).getPacketOffset();
			
			int eoff = pktCblkInfos.get(pktCblkInfos.size() -1).getPacketDataOffset() +
				pktCblkInfos.get(pktCblkInfos.size() -1).getPacketBodyLength();
			int elength = ehs.length() - eoff;
			
			int saveOff = ehs.getPos();
			
			mainHeader = new byte[hlength];
			epilog = new byte[elength];
			
			ehs.seek(0);
			ehs.readFully(mainHeader, 0, hlength);
			
			ehs.seek(eoff);
			ehs.readFully(epilog, 0, elength);
			
			
			
			ehs.seek(saveOff);
		} catch (IOException ioe ) {
			System.err.println ("Could not read main header: " + ioe.toString());
			System.exit(1);
		}
	}

	/**
	 * for each packet in the stream, read the related 
	 * packet header and packet body data and save it
	 * in the object
	 */
	public void readPacketData () {		

		Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			if (verbose) System.out.println ("Reading head of " + pci);
			pci.readHead(ehs);
			if (verbose) System.out.println ("Reading body of " + pci);
			pci.readBody(ehs);
		}
		
	}	
	
	/**
	 * Using the data collected previously, mess up the headers
	 * 
	 * @param maxRL the maximum resolution level
	 */
	public void mess() {
		System.out.println ("Forward transform of " + headersString() + ".");
		
		readMainHeadersAndEpilog();
		readPacketData();					
		mathematica = "1";
		
		if (verbose) 
			printInfo();
		
		metaRandom = new Random(seed);
		messInfoByRes();
		
		simulateBodyEncryption ();
		
		int tlen = createFakePacketHeaders();
		fakeMainHeader(tlen);
		writeBitstream("header-encrypted.j2k");
		FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Encrypted file written to header-encrypted.j2k.");
		
				
		if (verbose) System.out.println("\n\n N[Log[10,\n" + mathematica + "\n]]\n");
	}
	
	/** simulate body encryption (only for the written encrypted bitstream)
	 *
	 */
	private void simulateBodyEncryption () {
		Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();				
			if (pci.getPacketIdx() <= encBodyStopPIdx) {
				pci.simulateBodyEncryption();				
			}
				
		}
	}
	
	/**
	 * Decrypt previously transformed headers 
	 * 
	 * @param maxRL maximum resolution level
	 */
	public void unMess () {
		
		System.out.println ("Reverse transform of " + headersString() + ".");
		
		// save cblk states
		Iterator<PktCblksInfo> i = dc.getPktCblkInfos().iterator();
		while (i.hasNext()) {
			PktCblksInfo pci = i.next();
			pci.saveCblkState();
		}
		
		this.decode = true;
		metaRandom = new Random(seed);		
		unMessInfoByRes();	
		
	}
	

	/**
	 * mess up header information by resolution
	 *
	 */
	private void messInfoByRes() {		
		
		for (int r = startRL; r <= maxRL; r++) {
			Vector<PktCblksInfo> packetsInRes = dc.getPacketsInResolution(r);
			if (verbose) System.out.println(" ********** PACKETS IN  RESOLUTION " + r
					+ " ********** ");
			mathematica += "\n(* Resolution " + r + "*)\n";
			Iterator<PktCblksInfo> it = packetsInRes.iterator();
			while (it.hasNext()) {
				PktCblksInfo pci = it.next();
				pci.saveCblkState();
				if (verbose) System.out.println(pci.toString());
			}
			
			long seeds [] = new long[4];
			
			for (int i = 0; i < seeds.length; i++) {
				seeds[i] = metaRandom.nextLong();
			}
							
			if (headers.contains("m")) {
				random = new Random(seeds[0]);
				messSkippedMSBs(r);
			}
			//System.out.println (" Max LZB: " + maxLZB);
			
			if (headers.contains("c")) {
				random = new Random (seeds[1]);					
				messCodingPasses(r);
			}
			
			if (headers.contains("l")) {
				random = new Random (seeds[2]);
				messLenghts(r);
			}
			
			if (headers.contains("i")) {
				// need to save cblk state again
				// because permuteInclusionInfo uses it!
				it = packetsInRes.iterator();
				while (it.hasNext()) {
					PktCblksInfo pci = it.next();
					pci.saveCblkState();
				}

				random = new Random (seeds[3]);
				permuteInclusionInfo(r);
			}

			
		}
	}
	
	/** 
	 * restore header information by resolution
	 *
	 */
	private void unMessInfoByRes() {
		for (int r = startRL; r <= maxRL; r++) {
			Vector<PktCblksInfo> packetsInRes = dc.getPacketsInResolution(r);
			if (verbose) System.out.println(" ********** PACKETS IN  RESOLUTION " + r
					+ " ********** ");
			mathematica += "\n(* Resolution " + r + "*)\n";
			Iterator<PktCblksInfo> it = packetsInRes.iterator();
			while (it.hasNext()) {
				PktCblksInfo pci = it.next();
				if (verbose) System.out.println(pci.toString());
			}

			long seeds [] = new long[4];
			
			for (int i = 0; i < seeds.length; i++) {
				seeds[i] = metaRandom.nextLong();
			}
			
			if (headers.contains("i")) {
				random = new Random (seeds[3]);
				permuteInclusionInfo(r);
			}
			
			if (headers.contains("l")) {
				random = new Random (seeds[2]);
				messLenghts(r);
			}

			if (headers.contains("c")) {
				random = new Random (seeds[1]);					
				messCodingPasses(r);
			}

			if (headers.contains("m")) {
				random = new Random (seeds[0]);
				messSkippedMSBs(r);
			}
			
		}
	}
		
	/**
	 * update the length of the tile-part header to reflect the new headers!
	 * @param nlen the new length to be set
	 */
	public void fakeMainHeader (int nlen) {
		if (verbose) System.out.println ("Faking Main Header");
		fakeTileLength(nlen);
		fakeCOD();
	}
	
	private void fakeCOD () {
		boolean foundCOD =false;
		boolean foundFF = false;
		int i = 0;
		while (i < mainHeader.length && !foundCOD) {
			int b = mainHeader[i] & 0xFF;
			
			if ( b == 0xFF ) {
				foundFF = true;
				if (verbose) System.out.println ("found FF at " + i);
			}
			if (foundFF && (b == 0x52)) {
				if (verbose) System.out.println ("Next after 0xFF " + Integer.toHexString(b));
				if (verbose) System.out.println ("found COD at " + i);
				foundCOD = true;
			}
			i++;
		}
		//SKIP LCOD Paramter
		i+=2;
		
		int cstyle = mainHeader[i] & 0xFF;

		String info = "";		
		if ( (cstyle &  HeaderDecoder.SCOX_USE_SOP) != 0) {
			info += "SOP";
		};
		if ( (cstyle &  HeaderDecoder.SCOX_USE_EPH) != 0) {
			info += (info != "" ? ", " : "") + "EPH";
		};
		
		if (verbose) System.out.println ("Host Stream used markers for: " + (info != "" ? info : "[none]") + ".");
 
		if (sop)
			cstyle |= HeaderDecoder.SCOX_USE_SOP;
		else
			cstyle &= ~(HeaderDecoder.SCOX_USE_SOP);
		if (eph)
			cstyle |= HeaderDecoder.SCOX_USE_EPH;
		else			
			cstyle &= ~(HeaderDecoder.SCOX_USE_EPH);
		
		mainHeader[i] = (byte) cstyle;		
	}
	
	
	private void fakeTileLength  (int nlen) {
		boolean foundFF = false;
		boolean foundSOT = false;


		int SOTPos = 0;

		int i = 0;
		while (i < mainHeader.length && ! foundSOT) {
			int b = mainHeader[i] & 0xFF;			
			//System.out.println (Integer.toBinaryString(b));
			//System.out.println (Integer.toBinaryString(0xFF));
			//System.out.println (Integer.toBinaryString(0x90));

			if ( b == 0xFF ) {
				foundFF = true;
				if (verbose) System.out.println ("found FF at " + i);
				SOTPos = i;
			}
			if (foundFF && (b == 0x90)) {
				if (verbose) System.out.println ("Next after 0xFF " + Integer.toHexString(b));
				if (verbose) System.out.println ("found SOT at " + i);
				foundSOT = true;
			}
			i++;
		}
		if (verbose) System.out.println ("Main header length " + mainHeader.length);

		// now come two lsot bytes
		//i+=2;
		int lsot = (mainHeader[i] << 8 | mainHeader[i+1]) & 0xFF;
		if (lsot != 10) {
			System.err.println ("Something wrong with the stream");
		}
		i+= 2;

		//now comes the tile index
		int tIdx = (mainHeader[i] << 8 | mainHeader[i+1]) & 0xFF;
		if (verbose) System.out.println ("tIdx = " + tIdx);
		i+=2;

		//now we are at the length!!
		//for (int j = 0; j < 4; j++) {
		//System.out.println (Integer.toBinaryString((mainHeader[i+j]  & 0xFF)));
		//}

		int len =  ( (mainHeader[i]&0xFF)  << 24) |
		( (mainHeader[i+1] & 0xFF)  << 16) |
		( (mainHeader[i+2] & 0xFF) << 8) |
		(mainHeader[i+3] & 0xFF)    ;

		if (verbose) System.out.println ("Original tlen = " + len);	


		int tlen = mainHeader.length - SOTPos  + nlen;
		if (verbose) System.out.println ("New len =       " + tlen);

		byte[] b = new byte[4];
		for (int j = 0; j < 4; j++) {
			b[j] = (byte) (  (tlen >> (24-j*8)) & 0xFF );
			mainHeader[i+j] = b[j];
			//System.out.println ("NEw Bytes " + Integer.toBinaryString( ((int) b[j]) & 0xFF  ));
		}
	}

	
	/**
	 * for each packet create a new header from the previously transformed
	 * header information information
	 *   
	 * @return the sum of bytes in the headers and bodies of all packets 
	 */
	public int createFakePacketHeaders () {
		
		FakePktEncoder fakeEncoder = new FakePktEncoder(maxRL, this);
		int hlen = 0;
		int blen = 0;
		Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
		while (it.hasNext()) {
			PktCblksInfo pci = it.next();
			BitOutputBuffer hbuf = null;

			/*if (pci.getResolution() == 4 && pci.getLayer()==4) {
				System.out.println ("Evil packet");
				System.out.println (pci.getAllCBlks().get(0).ntp[pci.getLayer()]);
				System.out.println ("pci " + pci.toString());
			}*/

			hbuf = fakeEncoder.encodePacket(pci, hbuf);
			hlen += hbuf.getLength();
			
			//add space for markes if used
	        hlen += 
	        	(sop?Markers.SOP_LENGTH:0)  + 
	        	(eph?Markers.EPH_LENGTH:0);
						
			blen += pci.getPacketBodyLength();
			//System.out.println ("Wrote packet r=" + pci.getResolution() + " l= "+pci.getLayer());
			//System.out.println ("hbuf =" + hbuf.toBitString());
			byte[] hbarray = null;
			hbarray = hbuf.toByteArray(hbarray);
			pci.setFakedHeader(hbarray);		
		}

		return hlen+blen; //return tile-length
	}
	
	/**
	 * Write the packets with their transformed headers to file <tt>filename</tt> 
	 * 
	 * @param filename 
	 */
	public void writeBitstream (String filename) {
		try {
			FileOutputStream fos = new FileOutputStream(filename);
			fos.write(mainHeader);
			if (verbose) System.out.println ("Main header written, length is " + mainHeader.length);
			FileCodestreamWriter writer = new FileCodestreamWriter(fos, Integer.MAX_VALUE);
			Iterator<PktCblksInfo> it = dc.getPktCblkInfos().iterator();
			while (it.hasNext()) {
				PktCblksInfo pci = it.next();				
				pci.writeFakedPacket(writer, sop, eph);
			}
			if (verbose) System.out.println ("packet data " + writer.getLength());
			fos.write(epilog);
			if (verbose) System.out.println ("epilog is of length " + epilog.length);
			
			fos.close();
			
			System.out.println ("Total file size = " + ( mainHeader.length + writer.getLength() + epilog.length));
			
		} catch (IOException ioe) {			
			System.err.println ("could not create filecodestreamwriter: error: " + ioe.toString());
			System.exit(1);
		}
	}
	
	
	/**
	 * mess up inclusion info for res. r
	 * @param r resolution level
	 */	
	private void permuteInclusionInfo(int r) {
		if (verbose) System.out.println("Messing up inclusion info");
		Iterator<PktCblksInfo> it = dc.getPacketsInResolution(r).iterator();
		PktCblksInfo pci = null;
		PktCblksInfo pci_previous = null;

		// Fast forward to first candidate packet
		while (it.hasNext()) {
			pci = it.next();
			if (// this is implied: pci.getPacketBodyLength() > 0 && // need
				// data

			// to start I need at leas 1 non-empty and 1 empty cblk
			// otherwise there is nothing to exchange
			// packets with all FI or all PI are useless,
			// as i am permuting inclusion info
			pci.getEmptyCblks().size() > 0 && pci.getNonEmptyCblks().size() > 0) {
				break; // found my candidate!
			}
		}

		if (!it.hasNext()) {
			if (verbose) System.out.println("Resolution " + r
					+ " cannot be messed up. Too bad...");
			return;
		}

		 
		//Initial messup!
		if (pci.getPacketIdx() >= startIdx)
			initialMessUp(pci);

		while (it.hasNext()) {
			pci_previous = pci;
			pci = it.next();
			// mess up pci
			if (pci.getPacketIdx() >= startIdx)
				messUpPacket(pci, pci_previous);
		}
	}

	/**
	 * transform initial packet (first candidate found in stream) 
	 * 
	 * @param pci packet to transform 
	 */
	private void initialMessUp(PktCblksInfo pci) {
		if (verbose) System.out.println("Initial messup of " + pci);
		Vector<InclusionInfo> inc = pci.getInclusionVector();
		if (verbose) System.out.println("Read inclusion vector        : " + inc.toString());

		if (!decode) {
			Shuffler.shuffle(inc, random);
			if (verbose) System.out.println("Shuffled inclusion vector  :" + inc.toString());
		} else {
			Shuffler.unshuffle(inc, random);
			if (verbose) System.out.println("Unshuffled inclusion vector  :"
					+ inc.toString());
		}

		int numTot = pci.getAllCBlks().size();
		int numFI = pci.getCBlksByInclusion(InclusionInfo.FI).size();
		int numNI = pci.getCBlksByInclusion(InclusionInfo.NI).size();
		int numPI = pci.getCBlksByInclusion(InclusionInfo.PI).size();
		mathematica += "*Binomial[" + numTot + ", " + numFI + "]*Binomial["
				+ (numTot - numFI) + ", " + numNI + "]*Binomial["
				+ (numTot - numFI - numNI) + "," + numPI + "]";

		pci.setInclusionVector(inc, decode);
	}

	/**
	 * transform subsequenct packet 
	 * 
	 * @param pci the packet to transform
	 * @param pci_previous the previous packet
	 */
	private void messUpPacket(PktCblksInfo pci, PktCblksInfo pci_previous) {

		if (verbose) System.out.println("Messing up " + pci);

		// the sequence of FI and NI inclusions
		Vector<InclusionInfo> fini = pci.getInclusionVectorFINI();

		// the sequence of PI and PN inclusions
		Vector<InclusionInfo> pipn = pci.getInclusionVectorPIPN();

		// a copy of the original inclusion vector
		Vector<InclusionInfo> org = pci.getInclusionVector();

		// we will use this copy of the inclusion vector to mess wioth
		Vector<InclusionInfo> inc = pci.getInclusionVector();

		// save the original as a reference for debugging
		// later on
		pci.setSavedInclusionVector(org);
		

		// get the previous inclusion vector to know the boundaries of our mess
		Vector<InclusionInfo> inc_prev = pci_previous.getInclusionVector();

		// the previous and current inclusion vector should have the same size
		if (inc.size() != inc_prev.size()) {
			System.err.println("Something has gone very wrong here.");
			System.exit(1);
		}

		if (verbose) System.out.println(
				" #prev NI     : "+ pci_previous.getCBlksByInclusion(InclusionInfo.NI).size() + "\n" +
				" #fi          : "+ pci.getCBlksByInclusion(InclusionInfo.FI).size() + "\n"	+ 
				" #ni          : "	+ pci.getCBlksByInclusion(InclusionInfo.NI).size() + "\n" + 
				" #prev FIPIPN : "	+ (pci_previous.getAllCBlks().size() - pci_previous.getCBlksByInclusion(InclusionInfo.NI).size()) + "\n" + 
				" #pi          : "	+ pci.getCBlksByInclusion(InclusionInfo.PI).size() + "\n" +
				" #pn          : "	+ pci.getCBlksByInclusion(InclusionInfo.PN).size());

		mathematica += "* Binomial["+ pci_previous.getCBlksByInclusion(InclusionInfo.NI).size()	+ ", "
				+ pci.getCBlksByInclusion(InclusionInfo.FI).size()	+ "] * Binomial["
				+ (pci_previous.getAllCBlks().size() - pci_previous.getCBlksByInclusion(InclusionInfo.NI).size()) + ", "
				+ pci.getCBlksByInclusion(InclusionInfo.PI).size() + "]";

		// System.out.println (" This is fini: " + fini.toString());
		// System.out.println (" This is pipn: " + pipn.toString());

		// depending on whether we encode or decode we have to use shuffle or
		// unshuffle
		if (!decode) {
			Shuffler.shuffle(fini, random);
			Shuffler.shuffle(pipn, random);
		} else {
			Shuffler.unshuffle(fini, random);
			Shuffler.unshuffle(pipn, random);
		}

		for (int i = 0; i < inc_prev.size(); i++) {

			if (inc_prev.get(i) == InclusionInfo.NI) {
				// we can put our FI and NI here
				inc.set(i, fini.remove(0));
			} else {
				inc.set(i, pipn.remove(0));
			}
		}

		if (fini.size() > 0 || pipn.size() > 0) {
			System.err.println("Something has gone very wrong here.");
			System.exit(1);
		}

		// make the shuffled/unshuffled inclusion information real
		pci.setInclusionVector(inc, decode);

		if (verbose) {
			System.out.println("Previous inclusion vector: " + inc_prev.toString());
			System.out.println("New inclusion            : " + inc.toString());
			System.out.println("After setting incInfo    : " + pci.getInclusionVector().toString());
			if (!decode)
				System.out.println("Real inclusion vector    : " + org.toString());
			else {
				System.out.println("Read inclusion vector    : " + org.toString());
			}
		}
	}
	
	/**
	 * Transform the number of leading zero bitplanes for resolution r 
	 * @param r resolution level
	 */
	private void messSkippedMSBs (int r) {
				
		Vector<CBlkInfo> cblks = dc.getCodeBlocksInRes(r, startIdx);
		Collections.sort(cblks);
		Iterator<CBlkInfo> it = cblks.iterator();
		
		long seeds [] = new long [1];
		//create 3 new random generators!
		for (int j = 0; j < seeds.length; j++) {
			seeds[j] = random.nextLong();
		}
		
		Random localRandom;		
		localRandom = new Random (seeds[0]);		
		
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (!ccb.isEmpty()) {
				//if (ccb.msbSkipped > maxLZB) maxLZB = ccb.msbSkipped;
				int mod = localRandom.nextInt(256);
				if (decode) mod = -mod;
				ccb.msbSkipped = (ccb.msbSkipped + mod) % MAXBP;
				if (ccb.msbSkipped <= 0) ccb.msbSkipped += MAXBP;
			} else {
				System.err.println("Codeblock should not be empty!");
				System.exit(1);
			}
		}	
	}

		
	/**
	 * Transform the number of leading zero bitplanes for resolution r 
	 * @param r resolution level
	 */
	private void messSkippedMSBsOLD (int r) {
		
		Vector<Integer> lzbInfo = new Vector<Integer>();
		Vector<CBlkInfo> cblks = dc.getCodeBlocksInRes(r);
		Collections.sort(cblks);
		Iterator<CBlkInfo> it = cblks.iterator();
		
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (!ccb.isEmpty()) {
				lzbInfo.add(new Integer(ccb.msbSkipped));
			}
		}

		if (verbose) System.out.println ("lzbinfo\n" + lzbInfo);
				
		long seeds [] = new long [3];
		//create 3 new random generators!
		for (int j = 0; j < seeds.length; j++) {
			seeds[j] = random.nextLong();
		}
		
		Random localRandom;
		
		/*random = new Random (seeds[0]);		
		 it = all_cblks.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			int mod = random.nextInt(5);
			if (decode) mod = -mod;
			ccb.msbSkipped += mod;

		}
		
		if (true) return;*/	
				
		if (!decode) {
			localRandom = new Random (seeds[0]); 
			Shuffler.shuffle(lzbInfo, localRandom);
			localRandom = new Random (seeds[1]);
			Shuffler.redistributePreservingSumAllowZeros(lzbInfo, localRandom);
			localRandom = new Random (seeds[2]);
			Shuffler.shuffle(lzbInfo, localRandom);
		} else {
			localRandom= new Random (seeds[2]); 
			Shuffler.unshuffle(lzbInfo, localRandom);
			localRandom = new Random (seeds[1]);
			Shuffler.unRedistributePreservingSumAllowZeros(lzbInfo, localRandom);
			localRandom = new Random (seeds[0]);
			Shuffler.unshuffle(lzbInfo, localRandom);			
		}
		
		if (verbose) System.out.println ("shuffled\n" + lzbInfo);
		
		it = cblks.iterator();
		Iterator <Integer> jt = lzbInfo.iterator();
		while (it.hasNext()) {
			CBlkInfo ccb = it.next();
			if (! ccb.isEmpty()) {
				int lzb = jt.next();
				ccb.msbSkipped = lzb;
			}
		}		
	}
	
	/**
	 * use this to unmess skippedMSB with during the testing phase
	 *
	 */
	private void unMessSkippedMSBsEncoderEncrypted () {
		System.exit(1);
		/*System.out.println("UnMessing skipped MSBs");

		int count = 0;

		Collections.sort (allNonEmptyCblks);
		Iterator<CBlkInfo> fi = allNonEmptyCblks.iterator();

		while (fi.hasNext()) {
			CBlkInfo ccb = fi.next();
			count++;

			// System.out.println ("Old CCB: " + ccb.toString() + "
			// msbSkipped=" + ccb.msbSkipped);

			int mod = random.nextInt(12);
			//if (random.nextBoolean())
					//mod = -mod;


			ccb.msbSkipped -= mod;

			System.out.println ("Unmessed "  + ccb +  " mod=" + mod);
			// System.out.println ("New CCB: " + ccb.toString() + "
			// msbSkipped=" + ccb.msbSkipped);
			ccb.setFake(true);
		}		
		System.out.println("UnMessed  " + count + " codeblocks");
		*/
	}

	/**
	 * transform the number of coding passes for resolution r
	 * @param r resolution level
	 */
	private void messCodingPasses (int r) {
		if (verbose) System.out.println("Messing up the coding passes");
		
		/** the random generator(s) used locally, 
		 * seeded by <tt>random</tt> */ 
		Random localRandom;
		
		
		Iterator<PktCblksInfo> i = dc.getPacketsInResolution(r, startIdx).iterator();
		while (i.hasNext()) {
			PktCblksInfo pci = i.next();
			int layer = pci.getLayer();
			
			int totalLen = pci.getPacketBodyLength();
			
			// the number of packet borders
			int numBorders = pci.getNonEmptyCblks().size() -1;
									
			if (totalLen == 0 || numBorders <= 0) {
				if (verbose) System.out.println ("Cannot mess up passes for " + pci);
				continue;
			} else {
				if (verbose) System.out.println ("Messing up passes for " + pci);
			}
			Vector<Integer> passes= new Vector<Integer>();
			Vector<CBlkInfo> nec = pci.getNonEmptyCblks();
			Iterator<CBlkInfo> it = nec.iterator();
			while (it.hasNext()) {
				CBlkInfo ccb = it.next(); 
				passes.add(new Integer(ccb.ntp[layer]));
			}
			if (verbose) {
				System.out.println ("PCI                   :" + pci);
				System.out.println ("Number of Borders     :" + numBorders);
				System.out.println ("Total Packet Length   :" + totalLen);
			}
			
			
			if (verbose) System.out.println ("Original     : " + passes);
			
			long seeds [] = new long [3];
			//create 3 new random generators!
			for (int j = 0; j < seeds.length; j++) {
				seeds[j] = random.nextLong();
			}
			
			if (!decode) {
				localRandom = new Random(seeds[0]);
				Shuffler.shuffle(passes, localRandom);
				localRandom = new Random(seeds[1]);
				Shuffler.redistributePreservingSum(passes, localRandom);
				localRandom = new Random(seeds[2]);
				Shuffler.shuffle(passes, localRandom);
			} else {
				localRandom = new Random(seeds[2]);
				Shuffler.unshuffle(passes, localRandom);
				localRandom = new Random(seeds[1]);
				Shuffler.unRedistributePreservingSum(passes, localRandom);
				localRandom = new Random(seeds[0]);
				Shuffler.unshuffle(passes, localRandom);			
			}
			
			if (verbose) System.out.println ("Shuffled     : " + passes);
			
			it = nec.iterator();
			Iterator<Integer> lit = passes.iterator();			
			int sum = 0;
			int sum_org = 0;
			while (it.hasNext()) {
				CBlkInfo ccb = it.next(); 
				if (!lit.hasNext()) {
					System.err.println ("Something went severly wrong. Sorry.");
					System.exit(1);
				}
				Integer pass = lit.next();
				sum_org += ccb.ntp[layer];
				ccb.ntp[layer] = pass;
				ccb.updateCTP();
				sum += pass; 
			}
			if (verbose) System.out.println ("Sum " + sum  + " original= " + sum_org);
			
		}
		
	}

	/** 
	 * transform the packet lengths for resolution r
	 * @param r resolution level
	 */
	private void messLenghts(int r) {
		
		/** the random generator(s) used locally, 
		 * seeded by <tt>random</tt> */ 
		Random localRandom;
		
		if (verbose) System.out.println("Messing up the block-lengths");
		// *** Shuffle lengths ***/
		Iterator<PktCblksInfo> i = dc.getPacketsInResolution(r, startIdx).iterator();
		while (i.hasNext()) {
			PktCblksInfo pci = i.next();
			int layer = pci.getLayer();
			
			int totalLen = pci.getPacketBodyLength();
			
			// the number of packet borders
			int numBorders = pci.getNonEmptyCblks().size() -1;
									
			if (totalLen == 0 || numBorders <= 0) {
				if (verbose) System.out.println ("Cannot mess up lengths for " + pci);
				continue;
			} else {
				if (verbose) System.out.println ("Messing up lengths for " + pci);
			}
			Vector<Integer> lengths= new Vector<Integer>();
			Vector<CBlkInfo> nec = pci.getNonEmptyCblks();
			Iterator<CBlkInfo> it = nec.iterator();
			while (it.hasNext()) {
				CBlkInfo ccb = it.next(); 
				lengths.add(new Integer(ccb.len[layer]));
			}
			if (verbose) {
				System.out.println ("PCI                   :" + pci);		
				System.out.println ("Number of Borders     :" + numBorders);
				System.out.println ("Total Packet Length   :" + totalLen);
				System.out.println ("Original     : " + lengths);
			}			
			
			long seeds [] = new long [3];
			//create 3 new random generators!
			for (int j = 0; j < seeds.length; j++) {
				seeds[j] = random.nextLong();
			}
			
			if (!decode) {
				localRandom = new Random(seeds[0]);
				Shuffler.shuffle(lengths, localRandom);
				localRandom = new Random(seeds[1]);
				Shuffler.redistributePreservingSum(lengths, localRandom);
				localRandom = new Random(seeds[2]);
				Shuffler.shuffle(lengths, localRandom);
			} else {
				localRandom = new Random(seeds[2]);
				Shuffler.unshuffle(lengths, localRandom);
				localRandom = new Random(seeds[1]);
				Shuffler.unRedistributePreservingSum(lengths, localRandom);
				localRandom = new Random(seeds[0]);
				Shuffler.unshuffle(lengths, localRandom);			
			}
			if (verbose) System.out.println ("(un)Shuffled     : " + lengths);
			
			it = nec.iterator();
			Iterator<Integer> lit = lengths.iterator();			
			int sum = 0;
			int curOff = pci.getPacketDataOffset();
			while (it.hasNext()) {
				CBlkInfo ccb = it.next(); 
				if (!lit.hasNext()) {
					System.err.println ("Something went severly wrong. Sorry.");
					System.exit(1);
				}
				Integer length = lit.next();
				ccb.len[layer] = length;
				ccb.off[layer] = curOff;
				curOff += length;				
				sum += length; 
			}
			if (verbose) System.out.println ("Sum " + sum + " original length  " + totalLen);			
		}
	}
	

	/**
	 * print information on the collected packet-cblk objects
	 *
	 */
	public void printInfo() {
		System.out.println("Collected Info: ");
		long sumHL = 0;
		long sumBL = 0;
		
		readPacketData();

		Iterator<PktCblksInfo> i = dc.getPktCblkInfos().iterator();
		while (i.hasNext()) {
			PktCblksInfo pci = i.next();

			System.out.println(" PACKET " + pci.getPacketIdx() + " r="
					+ pci.getResolution() + " l=" + pci.getLayer());
			System.out.println("     offset=" + pci.getPacketOffset()
					+ " doff=" + pci.getPacketDataOffset() + " hlen="
					+ pci.getHeaderLength() + " bodysize="
					+ pci.getPacketBodyLength());
			
			sumHL += pci.getHeaderLength();
			sumBL += pci.getPacketBodyLength();
			
			System.out.print("      # codeblocks included for the first time: "
					+ pci.getCBlksByInclusion(InclusionInfo.FI).size());
			System.out.println(" SUM = "
					+ pci.getDataLenByInclusion(InclusionInfo.FI));

			System.out
					.print("      # codeblocks previously included and included: "
							+ pci.getCBlksByInclusion(InclusionInfo.PI).size());
			System.out.println(" SUM = "
					+ pci.getDataLenByInclusion(InclusionInfo.PI));

			System.out
				.println("      # codeblocks previously included and not included: "
							+ pci.getCBlksByInclusion(InclusionInfo.PN).size());
			System.out
					.println("      # codeblocks not included and not previously included: "				
							+ pci.getCBlksByInclusion(InclusionInfo.NI).size());
			//System.out
			//		.println ("  Header        : " + ArrayUtil.printBitArray(pci.getOriginalHeader()));
			
			//print decoded header (bitstuffing undone)
			System.out
					.print ("  Decoded Header: " );
			byte[] header = pci.getOriginalHeader();
			PktHeaderBitReader br = new PktHeaderBitReader(
					new ByteArrayInputStream(header));
			while (br.bais.available() > 0) {
				try {
					for (int b = 0; b <8; b++) {
						int bit = br.readBit();
						System.out.print (bit);
					}
					System.out.print (" ");
				} catch (IOException ioe) {
					System.out.println("IOE " + ioe.toString());
				}			
			} 
		}
		System.out.println();
		System.out.println (" Total Packet Header Data: " + sumHL);
		System.out.println (" Total Packet Body Data  : " + sumBL);

	}

	/** 
	 * is this messer in encoding or decoding mode?
	 * the decoding mode is set automatically by calling unmess
	 *  
	 * @return decoding mode
	 */
	public boolean isDecode() {
		return decode;
	}
	
	/**
	 * return as a string which parts of the headers are transfomed
	 * 
	 * @return parts of headers to be transformed as string
	 */
	private String headersString () {
		String hdrs = "";
		if (headers.contains("m")) {
			hdrs += "Skipped MSBs, ";
		}
		if (headers.contains("i")) {
			hdrs += "Inclusion Info, ";		
		}
		if (headers.contains("l")) {
			hdrs += "CCP Lenghts, ";
		}
		if (headers.contains("c")) {
			hdrs += "Coding Passes, ";
		}
		hdrs = hdrs.substring(0, hdrs.length()-2);
		hdrs += " (" + headers + ")";
		return hdrs;
	}
	
	/** get the DataCollector associated with this HeaderTransformer */
	public DataCollector getDataCollector () {
		return dc;
	}

}

/*   *** Some olde stuff: *** */
/*
	private void messInclusionInfoPI(int r) {
		System.out.println("Messing up inclusion info");
		System.out.println("Scanning for candidates");
		Iterator<PktCblksInfo> it = getPacketsInResolution(r).iterator();
		PktCblksInfo pci = null;
		while (it.hasNext()) {
			pci = it.next();

			Vector<CBlkInfo> pis = pci.getCBlksByInclusion(InclusionInfo.PI);
			Vector<CBlkInfo> pns = pci.getCBlksByInclusion(InclusionInfo.NI);

			if (pci.getPacketBodyLength() > 0 // need data
					&& pis.size() > 0 // need at least one previous inclusion
										// that is included here
					&& pns.size() > 0)// need at least one previous inclusion
										// that is not included here
			{
				// shift pci!
				int num_possible = Math.min(pis.size(), pns.size());

				System.out.println("Shifting " + num_possible + " blocks.");

				for (int i = 0; i < num_possible; i++) {
					CBlkInfo pii_ccb = pis.get(i);
					CBlkInfo pin_ccb = pns.get(i);
					int l = pci.getLayer();
					pii_ccb.setFake(true);
					pin_ccb.setFake(true);

					// make ni_ccb significant

					pin_ccb.ntp[l] = pii_ccb.ntp[l];
					pii_ccb.ntp[l] = 0;

					pin_ccb.off[l] = pii_ccb.off[l];
					pin_ccb.len[l] = pii_ccb.len[l];

					pii_ccb.off[l] = 0;
					pii_ccb.len[l] = 0;

					pin_ccb.updateCTP();
					pii_ccb.updateCTP();

					// the information regarding pktIdx is the same for
					// all cblks in a packet!
					// ni_ccb.pktIdx[l] = fi_ccb.pktIdx[l];
					// fi_ccb.pktIdx[l] = -1;
				}
			}
		}
	}

	private void messInclusionInfoFI(int r) {

		System.out.println("Messing up inclusion info");
		System.out.println("Scanning for first candidate");
		Iterator<PktCblksInfo> it = getPacketsInResolution(r).iterator();
		PktCblksInfo pci = null;
		while (it.hasNext()) {
			pci = it.next();
			Vector<CBlkInfo> fis = pci.getCBlksByInclusion(InclusionInfo.FI);
			Vector<CBlkInfo> nis = pci.getCBlksByInclusion(InclusionInfo.NI);

			if (pci.getPacketBodyLength() > 0 // need data
					&& fis.size() > 0 // need at least one FI
					&& nis.size() > 0)// need at least one not included to
										// swap

			{
				// shift pci!
				int num_possible = Math.min(fis.size(), nis.size());
				System.out.println("Shifting " + num_possible + " blocks.");

				for (int i = 0; i < num_possible; i++) {
					CBlkInfo fi_ccb = fis.get(i);
					CBlkInfo ni_ccb = nis.get(i);
					int l = pci.getLayer();
					fi_ccb.setFake(true);
					ni_ccb.setFake(true);

					// make ni_ccb significant

					ni_ccb.ntp[l] = fi_ccb.ntp[l];
					fi_ccb.ntp[l] = 0;

					ni_ccb.off[l] = fi_ccb.off[l];
					ni_ccb.len[l] = fi_ccb.len[l];

					fi_ccb.off[l] = 0;
					fi_ccb.len[l] = 0;

					ni_ccb.updateCTP();
					fi_ccb.updateCTP();

					// the information regarding pktIdx is the same for
					// all cblks in a packet!
					// ni_ccb.pktIdx[l] = fi_ccb.pktIdx[l];
					// fi_ccb.pktIdx[l] = -1;
				}
			}
		}

	}
	*/