/*
 * Created on Mar 16, 2007
 *
 */
package jj2000.j2k.encryption;

import java.util.Collections;
import java.util.List;
import java.util.ListIterator;
import java.util.Random;
import java.util.Vector;

public class Shuffler {
	
    /**
     * Randomly permute the specified list using the specified source of
     * randomness.  All permutations occur with equal likelihood
     * assuming that the source of randomness is fair.<p>
     *
     * This implementation traverses the list backwards, from the last element
     * up to the second, repeatedly swapping a randomly selected element into
     * the "current position".  Elements are randomly selected from the
     * portion of the list that runs from the first element to the current
     * position, inclusive.<p>
     *
     *
     * @param  list the list to be shuffled.
     * @param  rnd the source of randomness to use to shuffle the list.
     * @throws UnsupportedOperationException if the specified list or its
     *         list-iterator does not support the <tt>set</tt> operation.
     */
	public static void shuffle(List<?> list, Random rnd) {
		int size = list.size();
		for (int i=size; i>1; i--) {
			int idx = rnd.nextInt(i);			
			Collections.swap(list, i-1, idx);
		}

	}
	
	/** 
	 * Undo the shuffling on a list that has been done by <tt>shuffle</t>.
	 * 
	 * The same source of randomness has to be used.
	 * 
	 * @param list the list to be unshuffled
	 * @param rnd the source of randomness to use to unshuffle the list
	 * 
	 */
	public static void unshuffle(List<?> list, Random rnd) {
		int size = list.size();
		Vector<Integer> indices = new Vector<Integer>();
		for (int i = size; i>1; i--) {
			indices.add(rnd.nextInt(i));
		}
		Collections.reverse(indices);
		for (int i=2; i<=size; i++) {
			int idx = indices.get(i-2);
			Collections.swap(list, i-1, idx);
		}
	}
	
	/**
	 * 
	 * cross pairs of values, preserving the sum and never letting an element vanish
	 * all elements must be greater than 0!
	 *  
	 * @param list
	 * @param rnd
	 */
	public static void redistributePreservingSum (List<Integer> list, Random rnd) {
		int numBorders = list.size() - 1;
		if (numBorders <= 0) return;
				
		//each border can be changed
		for (int i = 0; i < numBorders; i++) {
			Integer el1 = list.get(i);
			Integer el2 = list.get(i+1);
			int sum = el1 + el2;
			if (sum == 0) continue; //ignore zeores
			float r = rnd.nextFloat();
			int mod = (int) (r*sum);
			int newBorder = (el1 + mod) % (sum-1) + 1;
			//System.out.println ("i=" + i + " m=" + mod + " nB="+newBorder+ " sum-1="+ (sum-1) );
			list.set (i, newBorder);
			list.set (i+1, sum-newBorder);			
			//System.out.println ( "   " + list);
		}
	}
	
	
	/**
	 * 
	 * cross pairs of values, preserving the sum and never letting an element vanish
	 * all elements must be greater than 0!
	 *  
	 * @param list
	 * @param rnd
	 */
	public static void unRedistributePreservingSum (List<Integer> list, Random rnd) {
		int numBorders = list.size() - 1;
		if (numBorders <= 0) return;
		
		Vector<Float> rndStream = new Vector<Float>();
		for (int i = 0; i < numBorders; i++) {			
			rndStream.add (rnd.nextFloat());
		}
		//Collections.reverse(rndStream);
		
		//traverse border backwards
		for (int i = numBorders - 1; i >= 0; i--) {
			Integer el1 = list.get(i);
			Integer el2 = list.get(i+1);
			int sum = el1 + el2;
			if (sum == 0) continue; //ignore zeores
			float r = rndStream.get(i);
			//System.out.println (r);
			int mod = (int) (r*sum);
			int newBorder = (el1 - mod -1) % (sum-1);
			if (newBorder <= 0) newBorder += (sum-1);	
			//System.out.println ("i=" + i + " m=" + mod + " nB="+newBorder + " sum-1="+ (sum-1) );
			list.set (i, newBorder);
			list.set (i+1, sum-newBorder);	
			//System.out.println ( "   " + list);
		}
				
	}
	
	/**
	 * 
	 * cross pairs of values, preserving the sum and never letting an element vanish
	 * all elements must be greater than 0!
	 *  
	 * @param list
	 * @param rnd
	 */
	public static void redistributePreservingSumAllowZeros (List<Integer> list, Random rnd) {
		int numBorders = list.size() - 1;
		if (numBorders <= 0) return;
				
		//each border can be changed
		for (int i = 0; i < numBorders; i++) {
			Integer el1 = list.get(i);
			Integer el2 = list.get(i+1);
			int sum = el1 + el2;
			if (sum == 0) continue; //ignore zero sums
			float r = rnd.nextFloat();
			int mod = (int) (r*sum);
			int newBorder = (el1 + mod) % (sum);
			//System.out.println ("i=" + i + " m=" + mod + " nB="+newBorder+ " sum-1="+ (sum-1) );
			list.set (i, newBorder);
			list.set (i+1, sum-newBorder);			
			//System.out.println ( "   " + list);
		}
	}
	
	
	/**
	 * 
	 * cross pairs of values, preserving the sum and never letting an element vanish
	 * all elements must be greater than 0!
	 *  
	 * @param list
	 * @param rnd
	 */
	public static void unRedistributePreservingSumAllowZeros (List<Integer> list, Random rnd) {
		int numBorders = list.size() - 1;
		if (numBorders <= 0) return;
		
		Vector<Float> rndStream = new Vector<Float>();
		for (int i = 0; i < numBorders; i++) {			
			rndStream.add (rnd.nextFloat());
		}
		//Collections.reverse(rndStream);
		
		//traverse border backwards
		for (int i = numBorders - 1; i >= 0; i--) {
			Integer el1 = list.get(i);
			Integer el2 = list.get(i+1);
			int sum = el1 + el2;
			if (sum == 0) continue; //ignore zeores
			float r = rndStream.get(i);
			//System.out.println (r);
			int mod = (int) (r*sum);
			int newBorder = (el1 - mod) % (sum);
			if (newBorder <= 0) newBorder += sum;	
			//System.out.println ("i=" + i + " m=" + mod + " nB="+newBorder + " sum-1="+ (sum-1) );
			list.set (i, newBorder);
			list.set (i+1, sum-newBorder);	
			//System.out.println ( "   " + list);
		}
				
	}


}
