/* 
 * CVS identifier:
 * 
 * $Id: Sobel.java,v 1.3 2005-11-15 13:58:34 dengel Exp $
 * 
 * Class:                   Sobel
 * 
 * Description:             
 * 
 * Created on Feb 7, 2005
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */
package jj2000.j2k.image.edgedectection;

import jj2000.j2k.image.DataBlkInt;

/**
 * Sobel Edge Operator
 *
 */
public class Sobel {

	/**
	 * the horizontal convolution kernel
	 */
	final static private int Gx[][] = {
			{-1, 0, 1},
			{-2, 0, 2},
			{-1, 0, 1}};
	
	/**
	 * the vertical convolution kernel 
	 */
	final static private int Gy[][] = {
			{ 1, 2, 1},
			{ 0, 0, 0},
			{-1,-2,-1}};
	
	/**
	 * 
	 * convolve kernel into image at position x, y
	 * 
	 * @param src
	 * @param kernel
	 * @param x
	 * @param y
	 * @return
	 */
	private int convolve (DataBlkInt src, int[][] kernel, int x, int y)
	{
		//convolute a gradient kernel in a pixel
		if (x == 0 || x == src.w - 1 
				|| y == 0 || y == src.h - 1) {
				return 0;
			}

		int offset = kernel.length / 2;
		int sum = 0;
		for (int j = 0; j < kernel.length; j++)
			for (int i = 0; i < kernel[j].length; i++) {							
				sum += kernel[i][j] * src.data[src.w*(y+j-offset) + x+i-offset];				
			}
		return sum;		
	}
	
	/**
	 * returns the gradient for a specified pixel position
	 * 
	 * @param src
	 * @param x
	 * @param y
	 * @return
	 */
	public double[] pixelGradient (DataBlkInt src, int x, int y)
	{
		int gx;
		int gy;
		 			
		gx = convolve(src, Gx, x, y);
		gy = convolve(src, Gy, x, y);
		
		double grad [] = {gx, gy};
		
		return grad;		
	}
		
	
	/**
	 * Calculates the average gradient for an image(-block)
	 * 
	 * @param blk
	 * @return
	 */
	public double[] imageGradient(DataBlkInt blk)
	{
		int w = blk.w;
		int h = blk.h;
		int ulx = blk.ulx;
		int uly = blk.uly;
		
		double sum_gx = 0;
		double sum_gy = 0;
		
		for (int j = 0; j < h; j++)
			for (int i =0; i < w; i++)
			{
				sum_gx += pixelGradient(blk, i, j)[0];
				sum_gy += pixelGradient(blk, i, j)[1];
			}
		
		double grad [] = {sum_gx / (h*w) , sum_gy / (h*w)};
		
		return grad;		
		
	}
	
	/**
	 * Calculates the dominant angle of lines for an image(-block)
	 * the vertical axes is 0, the value will increase for edges anticlockwise of this and
	 * decrease for edges clockwise of this, in a range of -pi/2 to pi/2
	 * 
	 * @param src
	 * @return
	 */
	public double imageAngle (DataBlkInt blk)
	{
		int w = blk.w;
		int h = blk.h;
		int ulx = blk.ulx;
		int uly = blk.uly;
		
		double[] grad = imageGradient(blk);
		double angle = 0;
	
		if (grad[0]== 0) {
			if (grad[1] == 0)
				angle = 0;
			else
				angle = Math.PI/2;
		}		
		else {
			angle = Math.atan((double) grad[1]/ (double) grad[0]);
		}
		return Math.toDegrees(angle);
	}	
	
	
	/**
	 * Get the id of the representative angle of an image(-block)
	 * images with gradient magnitude below threshold are considered
	 * non-edge images (id 0) 
	 * 
	 * @param blk
	 * @param threshold 
	 * @return id
	 */
	public int imageReprAngleID(DataBlkInt blk, double threshold)
	{
		
		
		double grad [] = imageGradient(blk);
		double mag =  Math.sqrt(Math.pow (grad[0], 2) + Math.pow(grad[1], 2));
		//System.out.println ("|Grad|= "  + mag);
		if (mag < (double) threshold)
		{
			return 0;
		}		
		
		double angle = imageAngle(blk);
		int id = 1 + (int) (8*(  (angle + 89.9) / 180));
		return id;
	}
	
	
	/**
	 * get the angle in degrees associated with a representative angle id
	 * 
	 * @param id
	 * @return angle
	 */
	public double IdtoReprAngl (int id)
	{		
		if ( id < 1 || id > 8)
		{
			throw new IllegalArgumentException("id out of range");
		}
		
		double angle = ((double) (id - 1))/8 * 180 - (90-11.25);
		
		return  angle;
	}
 
	
	/**
	 * 
	 * Returns an edge image constructed with the sobel operator
	 * 
	 * @param src
	 * @param threshold threshold to apply
	 * @return edge image
	 */
	public DataBlkInt edgeImage (DataBlkInt src, int threshold)
	{
		int w = src.w;
		int h = src.h;
		int ulx = src.ulx;
		int uly = src.uly;
		int gx;
		int gy;
		int grad;
		
		DataBlkInt edgeImg = new DataBlkInt(ulx, uly, w, h);
		
		for (int j = 0; j < h; j++)
			for (int i =0; i < w; i++)
			{
				gx = convolve(src, Gx, i, j);
				gy = convolve(src, Gy, i, j);
				grad = (int) Math.sqrt(Math.pow (gx, 2) + Math.pow (gy, 2));
				if (grad < threshold)
				{
					grad = 0;
				}
				edgeImg.data[j*w + i] = grad; 
			}		
		return edgeImg;		
	}
	
}
