/* 
 * CVS identifier:
 * 
 * $Id: FrequencyDomainAnalyzer.java,v 1.5 2008-05-05 09:25:04 dengel Exp $
 * 
 * Class:                   FrequencyDomainAnalyzer
 * 
 * Description:             
 * 
 * Created on June 16, 2005
 * 
 * @author Dominik Engel <dengel@cosy.sbg.ac.at>
 * 
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 * 
 * 
 * 
 */


package jj2000.j2k.wavelet;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

import jj2000.j2k.image.DataBlk;
import jj2000.j2k.image.DataBlkFloat;
import jj2000.j2k.image.DataBlkInt;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;
import jj2000.j2k.util.ParameterList;

/**
 * 
 * Analysis in the frequency Domain
 * 
 * @author dengel
 *  
 */
public class FrequencyDomainAnalyzer {

	public final static int WRITE_ASCII_PGM = 0;
	public final static int WRITE_ASCII = 1;
	public final static int WRITE_RAW = 2;
	
	/**
	 * this class represents a function to be iterated over a subband
	 *  
	 */
	private abstract class Function {

		/** The containter for accumulation */
		protected float res = 0.0f;

		/** reset the containter */
		protected void reset() {
			res = 0;
		}

		/** return result */
		protected float getResult() {
			return res;
		}

		/**
		 * apply the function to a coefficient, return result
		 * 
		 * @param val
		 */
		protected abstract float apply(float val);

		/**
		 * override this function to specify what should be done before
		 * iteration starts
		 * 
		 * @param sb
		 */
		protected void initializeIteration(Subband sb) {
		};

		/**
		 * apply function to subband, return accumulated value
		 * 
		 * @param sb
		 * @return
		 */
		public float applyToSubband(Subband sb) {
			initializeIteration(sb);
			for (int iy = 0; iy < sb.h; iy++)
				for (int ix = 0; ix < sb.w; ix++) {
					res += apply(coeffs[(sb.ulx + ix) * w + (sb.uly + iy)]);
				}
			float result = getResult();
			reset();
			return result;
		}

		/**
		 * Print evaluation for all subbands in the tree sb
		 * 
		 * @param sb
		 * @return
		 */
		public String printOverSubbandTree(Subband sbtree) {
			String sRes = "";
			float acc = 0;
			int count = 0;
			Subband sb = sbtree.getDeepestGlobalLL();
			while (sb != null) { // loop on res

				Subband sb2 = sb;
				while (sb2 != null) // loop in res
				{
					float res = applyToSubband(sb2);
					acc += Math.abs(res);

					count++;
					reset();

					sRes += sb2.resLvl + "/" + sb2.sbandIdx + ": " + res + "\n";

					sb2 = sb2.nextSubband();
				}

				sb = sb.getNextResLevel();

			}
			sRes += "  Average (abs): " + acc / count + "\n";
			return sRes;

		}

		/**
		 * Evaluate the function for each subband and write the result to a file
		 * First entry is the average for all subbands, seconde the avg of the 
		 * detail subbands
		 * 
		 * @param sbtree
		 * @return
		 */
		public void writeToGPFilePerSubband(Subband sbtree, String fname) {
			float acc = 0;
			float approx = 2; // some number not in [-1,1]
			float result = 0;
			String sres = " ";
			int count = 0;
			Subband sb = sbtree.getDeepestGlobalLL();
			while (sb != null) { // loop on res

				Subband sb2 = sb;
				while (sb2 != null) // loop in res
				{
					result = applyToSubband(sb2);
					if (approx == 2) {
						approx = result;
					}						
					acc += Math.abs(result);
					sres += " " + Float.toString(result);
					count++;
					reset();
					sb2 = sb2.nextSubband();
				}
				sb = sb.getNextResLevel();
			}
			GPwriteResult(fname, Float.toString(acc / count) + " "  
					+ Float.toString ( ((acc - Math.abs(approx)) / (count - 1)) ) + 
							sres, false);
		}

		/**
		 * 
		 * average function values over subband
		 * 
		 * @param sb
		 * @return
		 */
		public float absAverageOverSubbandTree(Subband sbtree) {

			float res = 0;
			int count = 0;
			Subband sb = sbtree.getDeepestGlobalLL();
			while (sb != null) { // loop on res

				Subband sb2 = sb;
				while (sb2 != null) // loop in res
				{
					res += Math.abs(applyToSubband(sb2));
					count++;
					reset();
					sb2 = sb2.nextSubband();
				}
				sb = sb.getNextResLevel();
			}
			return res / count;
		}
	}

	/**
	 * Energy
	 *  
	 */
	private class fEnergy extends Function {
		public float apply(float val) {
			return (float) Math.pow(val, 2);
		}
	}

	/**
	 * Mean
	 *  
	 */
	private class fMean extends Function {
		private int count = 0;

		public float getResult() {
			return res / count;
		}

		protected float apply(float val) {
			count++;
			return val;
		}
	}

	/**
	 * Variance
	 *  
	 */
	private class fVariance extends Function {
		private float mean = 0f;

		protected void initializeIteration(Subband sb) {
			this.mean = (new fMean()).applyToSubband(sb);
		}

		protected float apply(float val) {
			return (float) Math.pow(val - mean, 2);
		}
	}

	/**
	 * Co-Variance
	 *  
	 */
	private class fCoVariance extends Function {
		private float prevVal = 0f;

		private float mean = 0f;

		protected void initializeIteration(Subband sb) {
			this.mean = (new fMean()).applyToSubband(sb);
		}

		protected float apply(float val) {
			float result = 0f;
			if (prevVal != 0) {
				result += (prevVal - mean) * (val - mean);
			}
			prevVal = val;
			return result;
		}
	}

	/**
	 * Correlation Coefficient
	 *  
	 */
	private class fCorrelationCoefficient extends Function {

		protected float apply(float val) {
			return 0;
		}

		public float applyToSubband(Subband sb) {
			float cov = (new fCoVariance()).applyToSubband(sb);
			float var = (new fVariance()).applyToSubband(sb);

			res = cov / var;

			return res;
		}
	}

	/** the subband structure to be analyzed */
	private Subband subb;

	/** the coefficients to be analyzed */
	public float[] coeffs;

	/** height of the coefficient array */
	int h = 0;

	/** width of the coefficient array */
	int w = 0;
	
	/**
	 * Create an empty FDA for w*h
	 * 
	 * @param w
	 * @param h
	 */
	public FrequencyDomainAnalyzer(int w, int h) {

		this.w = w;
		this.h = h;
		coeffs = new float[h * w];
		subb = null;

	}

	

	/**
	 * Create a FDA for specified coefficient data with subband structure sb
	 * 
	 * @param sb
	 * @param coefficients
	 */
	public FrequencyDomainAnalyzer(Subband sb, DataBlk coefficients) {

		h = coefficients.h;
		w = coefficients.w;

		subb = sb;

		coeffs = new float[h * w];
		
		if (coefficients.getDataType() == DataBlk.TYPE_FLOAT) {
		// Copy data into own buffer, so we can mess with it
			readCoefficientsFromDataBlk((DataBlkFloat) coefficients);
		} else {
			readCoefficientsFromDataBlk((DataBlkInt) coefficients);
		}
	}
	

	
	/**
	 * copy coefficients from given DataBlk to internal buffer
	 * 
	 * @param coefficients DataBlk to be copied
	 */
	public void readCoefficientsFromDataBlk (DataBlkFloat coefficients)
	{
		System.arraycopy(coefficients.data, 0, coeffs, 0, h * w);
	}
	
	/**
	 * copy coefficients from given DataBlk to internal buffer
	 * 
	 * @param coefficients DataBlk to be copied
	 */
	public void readCoefficientsFromDataBlk (DataBlkInt coefficients)
	{
		for (int i = 0; i < h*w; i++) {
			coeffs[i] = (float) coefficients.data[i];
		}
	}
	
	
	
	/**
	 * write coefficients from internal buffer to given DataBlk
	 * 
	 * @param target DataBlk to write coefficients to
	 */
	public void writeCoefficientsToDataBlk (DataBlkFloat target)
	{
		System.arraycopy(coeffs, 0, target.data, 0, h * w);
	}
	
	/**
	 * write coefficients from internal buffer to given DataBlk
	 * 
	 * @param target DataBlk to write coefficients to
	 */
	public void writeCoefficientsToDataBlk (DataBlkInt target)
	{
		for (int i = 0; i < h*w; i++) {
			target.data[i] = (int) (coeffs[i]);
		}		
	}
	
	/**
	 * read a set of coefficients from file fname into private buffer
	 * 
	 * @param fname
	 */
	public void readCoefficientsFromFile (String fname)
	{	
		try {
			FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Reading Coefficients from \"" + fname + "\"");
			FileInputStream fis = new FileInputStream(fname);
			DataInputStream ins = new DataInputStream(fis);
					
			int i = 0;
			while (ins.available() != 0)
			{																				
				coeffs[i++] = ins.readFloat();
			}
		}
		catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, 
					"Could not open " + fname + ".");
			return;					
		}
		catch (IOException ioe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, 
					"IOException trying to read from " + fname + ".");
			return;
		}
	}
	
	/**
	 * Write energy of the coefficients to File
	 * 
	 * @param fname
	 *            filename
	 * @param with_coord
	 *            also include coordinates
	 * @param as_pgm
	 *            write pgm preamble, denormalize & round values
	 */
	public void writeEnergyToFile(String fname, boolean with_coord,
			int type) {
		try {
			FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Writing Energy Coefficients to \"" + fname + "\""
							+ ( type == WRITE_ASCII_PGM ? " (PGM)" : 
								type == WRITE_ASCII ? " (ascii)" : " (raw)"));
			FileOutputStream fos = new FileOutputStream(fname, false);
			DataOutputStream out = new DataOutputStream(fos);

			if (type == WRITE_ASCII_PGM) {
				out.writeBytes("P2\n" + w + " " + h + "\n255\n");
			}

			double max = 0;
			for (int iy = 0; iy < h; iy++)
				for (int ix = 0; ix < w; ix++) {
					double energy = Math.pow (coeffs[iy * w + ix], 2);
					if (energy > max)
						max = energy;					
			}
			

			//ASCII TYPES
			if (type != WRITE_RAW) {
			for (int iy = 0; iy < h; iy++)
				for (int ix = 0; ix < w; ix++) {
					String stmp = "";
					if (with_coord) {
						stmp += ix + " " + iy + " ";
					}
					if (type != WRITE_ASCII_PGM) { //no offset for ASCII_RAW
						stmp += coeffs[iy * w + ix] + "\n";
					} else { //offset for PGM
						double energy = Math.pow (coeffs[iy * w + ix], 2);
						double val = (energy/max) * 255;
						stmp += Math.round(val)	+ "\n";
					}
					out.writeBytes(stmp); 
				}											
			} else { //RAW
				for (int i = 0; i < w*h; i++)
				{
					out.writeFloat(coeffs[i]);
				}
			}
			out.close();
		} catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
					"File not Found");
		} catch (IOException ioe) {
			FacilityManager.getMsgLogger()
					.printmsg(MsgLogger.ERROR, "IO Error");
		} catch (Exception e) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
					e.toString());
		}

	}


	/**
	 * Write all the coefficients to File
	 * 
	 * @param fname
	 *            filename
	 * @param with_coord
	 *            also include coordinates
	 * @param as_pgm
	 *            write pgm preamble, denormalize & round values
	 */
	public void writeCoefficientsToFile(String fname, boolean with_coord,
			int type) {
		try {
			FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Writing Coefficients to \"" + fname + "\""
							+ ( type == WRITE_ASCII_PGM ? " (PGM)" : 
								type == WRITE_ASCII ? " (ascii)" : " (raw)"));
			FileOutputStream fos = new FileOutputStream(fname, false);
			DataOutputStream out = new DataOutputStream(fos);

			if (type == WRITE_ASCII_PGM) {
				out.writeBytes("P2\n" + w + " " + h + "\n255\n");
			}

			//ASCII TYPES
			if (type != WRITE_RAW) {
			for (int iy = 0; iy < h; iy++)
				for (int ix = 0; ix < w; ix++) {
					String stmp = "";
					if (with_coord) {
						stmp += ix + " " + iy + " ";
					}
					if (type != WRITE_ASCII_PGM) { //no offset for ASCII_RAW
						stmp += coeffs[iy * w + ix] + "\n";
					} else { //offset for PGM
						stmp += Math.abs(Math.round(coeffs[iy * w + ix] + 128))
								+ "\n";
					}
					out.writeBytes(stmp); 
				}											
			} else { //RAW
				for (int i = 0; i < w*h; i++)
				{
					out.writeFloat(coeffs[i]);
				}
			}
			out.close();
		} catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
					"File not Found");
		} catch (IOException ioe) {
			FacilityManager.getMsgLogger()
					.printmsg(MsgLogger.ERROR, "IO Error");
		} catch (Exception e) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
					e.toString());
		}

	}

	/** do the standard analysis, whatever this might be */
	public void handle(ParameterList pl) {
		
		if (pl.getBooleanParameter("Wfreqanalysis"))
		{
			//writeCoefficientsToFile("test.pgm", false, WRITE_ASCII_PGM);
			//writeEnergyToFile ("test.pgm", false, WRITE_ASCII_PGM);
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
					"Energy\n" + (new fEnergy()).printOverSubbandTree(subb));
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
					"Mean\n" + (new fMean()).printOverSubbandTree(subb));
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
					"Variance\n" + (new fVariance()).printOverSubbandTree(subb));
			FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Co-Variance\n"
							+ (new fCoVariance()).printOverSubbandTree(subb));
			FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Correlation Coefficient\n"
							+ (new fCorrelationCoefficient())
									.printOverSubbandTree(subb));
			/*FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Correlation Coefficient Avg "
							+ (new fCorrelationCoefficient())
									.absAverageOverSubbandTree(subb));*/
		}

		if (pl.getParameter("GPFile") != null) {
			if (pl.getParameter("GPStyle").equalsIgnoreCase("cor")) {
				(new fCorrelationCoefficient()).writeToGPFilePerSubband(subb,
						pl.getParameter("GPFile"));
			}
		}
		
		if (pl.containsKey("Wwritecoefffile"))
		{
			writeCoefficientsToFile(pl.getParameter("Wwritecoefffile"), false, WRITE_RAW);
		}
		
	}

	/**
	 * Write a single string value to a result file
	 * 
	 * @param fname
	 *            filename of the gnuplot file
	 * @param res
	 *            result value
	 * @param lf
	 *            include linefeed?
	 */
	public void GPwriteResult(String fname, String res, boolean lf) {
		try {

			FileOutputStream fos = new FileOutputStream(fname, true);
			DataOutputStream out = new DataOutputStream(fos);
			//if (res < 0) res = 0;
			//if (Double.isNaN(res)) res = 0;
			out.writeBytes(res + " " + (lf ? "\n" : ""));
			out.close();
		} catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
					"File not Found");
		} catch (IOException ioe) {
			FacilityManager.getMsgLogger()
					.printmsg(MsgLogger.ERROR, "IO Error");
		} finally {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
					"GP appended to " + fname);

		}

	}

}