/*
 * CVS identifier:
 *
 * $Id: SubbandAniso.java,v 1.1 2005-05-17 10:07:46 dengel Exp $
 *
 * Class:                   SubbandAniso
 *
 * Description:             Asbtract element for a tree structure for
 *                          a description of anisotropic subbands.
 *
 *
 *
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 *  */
package jj2000.j2k.wavelet;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;

import jj2000.j2k.util.ArrayUtil;
import jj2000.j2k.util.FacilityManager;
import jj2000.j2k.util.MsgLogger;

/**
 * This abstract class represents a subband in a bidirectional tree structure
 * that describes the subband decomposition for a wavelet transform. This class
 * is implemented by the SubbandAn and SubbandSyn classes, which are for the
 * analysis and synthesis sides, respectively.
 * 
 * <P>
 * The element can be either a node or a leaf of the tree. If it is a node then
 * ther are 4 descendants (LL, HL, LH and HH). If it is a leaf ther are no
 * descendants.
 * 
 * <P>
 * The tree is bidirectional. Each element in the tree structure has a "parent",
 * which is the subband from which the element was obtained by decomposition.
 * The only exception is the root element which has no parent (i.e.it's null),
 * for obvious reasons.
 * 
 * @see jj2000.j2k.wavelet.analysis.SubbandAn
 * @see jj2000.j2k.wavelet.synthesis.SubbandSyn
 */
public abstract class SubbandAniso {

	/** The ID for the LL orientation */
	public final static int WT_ORIENT_L = 0;

	/** The ID for the HL (horizontal high-pass) orientation */
	public final static int WT_ORIENT_H = 1;
	
	public final static int WT_DYADIC = 0;
	public final static int WT_FULL = 1;
	
	public final static int DIM_UNSPECIFIED = -1;
	public final static int DIM_HOR = 0;
	public final static int DIM_VER = 1;
	
	
	
	
	/**
	 * get high-pass child in dimension d
	 * corresponds to Subband.getXX
	 * @param d dimension
	 * @return
	 */
	public abstract SubbandAniso getH ();
	
	/**
	 * get low-pass child in dimension d
	 * corresponds to Subband.getXX
	 * @param d dimension
	 * @return
	 */
	public abstract SubbandAniso getL ();
	

	/**
	 * True if it is a node in the tree, false if it is a leaf. False by
	 * default.
	 */
	public boolean isNode;
	
	/**
	 * the dimension into which this node is _to be_ decomposed
	 * defaults to unspecified (which it is for leaves)
	 */
	public int dimension = DIM_UNSPECIFIED;

	
	/**
	 * The orientation of this subband (WT_ORIENT_LL, WT_ORIENT_HL,
	 * WT_ORIENT_LH, WT_ORIENT_HH). It is WT_ORIENT_LL by default. The
	 * orientation of the top-level node (i.e. the full image before any
	 * decomposition) is WT_ORIENT_LL.
	 */
	// The default value is always 0, which is WT_ORIENT_LL.
	public int orientation;

	/**
	 * The global orientation of the subband. This corresponds to the
	 * orientation of the first non-LL parent of the subband, going from the
	 * leaves to the root.
	 */
	// The default value is always 0, which is WT_ORIENT_LL.
	public int gOrient;

	/**
	 * The level in the tree to which this subband belongs, which is the number
	 * of wavelet decompositions performed to produce this subband. It is 0 for
	 * the top-level (i.e. root) node. It is 0 by default.
	 */
	public int level;

	/**
	 * The resolution level to which this subband contributes. Level 0 is the
	 * smallest resolution level (the one with the lowest frequency LL subband).
	 * It is 0 by default.
	 * Note that the resolution level does not necessarily coincide with the 
	 * number of wavelet decompositions needed to obtain this subband. The
	 * resolution level only gives information to which resolution of the 
	 * image this subband contributes
	 * 
	 * @see #level
	 */
	public int resLvl;

	/**
	 * The base 2 exponent of the analysis gain of the subband. The analysis
	 * gain of a subband is defined as the gain of the previous subband (i.e.
	 * the one from which this one was obtained) multiplied by the line gain and
	 * by the column gain. The line (column) gain is the gain of the line
	 * (column) filter that was used to obtain it, which is the DC gain for a
	 * low-pass filter and the Nyquist gain for a high-pass filter. It is 0 by
	 * default.
	 * 
	 * <P>
	 * Using the base 2 exponent of the value contrains the possible gains to
	 * powers of 2. However this is perfectly compatible to the filter
	 * normalization policy assumed here. See the split() method for more
	 * details.
	 * 
	 * @see #split
	 */
	public int anGainExp;

	/**
	 * The subband index within its resolution level. This value uniquely
	 * identifies a subband within a resolution level and a decomposition level
	 * within it. Note that only leaf elements represent "real" subbands, while
	 * node elements represent only intermediate stages.
	 * 
	 * <P>
	 * It is defined recursively. The root node gets a value of 0. For a given
	 * node, with a subband index 'b', its LL descendant gets 4*b, its HL
	 * descendant 4*b+1, its LH descendant 4*b+2, and its HH descendant 4*b+3,
	 * for their subband indexes.
	 */
	public int sbandIdx = 0;

	/**
	 * The horizontal coordinate of the upper-left corner of the subband, with
	 * respect to the canvas origin, in the component's grid and subband's
	 * decomposition level. This is the real horizontal index of the first
	 * column of this subband. If even the horizontal decomposition of this
	 * subband should be done with the low-pass-first convention. If odd it
	 * should be done with the high-pass-first convention.
	 */
	public int ulcx;

	/**
	 * The vertical coordinate of the upper-left corner of the subband, with
	 * respect to the canvas origin, in the component's grid and subband's
	 * decomposition level. This is the real vertical index of the first column
	 * of this subband. If even the vertical decomposition of this subband
	 * should be done with the low-pass-first convention. If odd it should be
	 * done with the high-pass-first convention.
	 */
	public int ulcy;

	/** The horizontal coordinate of the upper-left corner of the subband */
	public int ulx;

	/** The vertical coordinate of the upper-left corner of the subband */
	public int uly;

	/** The width of the subband */
	public int w;

	/** The height of the subband */
	public int h;

	/** The nominal code-block width */
	public int nomCBlkW;

	/** The nominal code-block height */
	public int nomCBlkH;

	/**
	 * Returns the parent of this subband. The parent of a subband is the
	 * subband from which this one was obtained by decomposition. The root
	 * element has no parent subband (null).
	 * 
	 * @return The parent subband, or null for the root one.
	 */
	public abstract SubbandAniso getParent();

	
	/**
	 * Splits the current subband in its four subbands. This creates the four
	 * childs (LL, HL, LH and HH) and converts the leaf in a node.
	 * 
	 * @param hfilter
	 *            The horizontal wavelet filter used to decompose this subband.
	 * 
	 * @param vfilter
	 *            The vertical wavelet filter used to decompose this subband.
	 * 
	 * @return A reference to the LL leaf (getLL()).
	 */
	public abstract SubbandAniso split(int dimension, WaveletFilter hfilter,
			WaveletFilter vfilter);
	
	/**
	 * 
	 * Take away all children, make node a leaf
	 *
	 */
	public void bereave () {
		isNode = false;
	}

	/**
	 * Initializes the childs of this node with the correct values. The sizes of
	 * the child subbands are calculated by taking into account the position of
	 * the subband in the canvas.
	 * 
	 * <P>
	 * For the analysis subband gain calculation it is assumed that analysis
	 * filters are normalized with a DC gain of 1 and a Nyquist gain of 2.
	 */
	protected void initChilds() {
		SubbandAniso subb_L = getL();
		SubbandAniso subb_H = getH();

		if (dimension == DIM_HOR)
		{
			subb_L.orientation = WT_ORIENT_L;
			//subb_L.dimension = dim; UNSPEC!!
			subb_L.level = level + 1;
			subb_L.ulx = ulx;
			subb_L.uly = uly;			
			subb_L.w = (w + 1) >> 1;
			subb_L.h = h;
			subb_L.gOrient = gOrient;
			subb_L.resLvl = (gOrient == WT_ORIENT_L) ? resLvl - 1 : resLvl;
			subb_L.anGainExp = anGainExp;
			subb_L.sbandIdx = (sbandIdx << 1); //R_0 will stay 0 		
			
			
			subb_H.orientation = WT_ORIENT_H;
			//subb_H.dimension = dim; !! UNSPEC
			subb_H.gOrient = (gOrient == WT_ORIENT_L) ? WT_ORIENT_H : gOrient;
			subb_H.level = subb_L.level;
			subb_H.ulx = ulx + subb_L.w;
			subb_H.uly = uly;
			subb_H.w = subb_L.w;
			subb_H.h = subb_L.h;
			subb_H.resLvl = resLvl;
			//gain 1 for highpass ==> lp + 0, lh + 1, hl + 1, hh + 2
			subb_H.anGainExp = anGainExp + 1; 
			subb_H.sbandIdx = (sbandIdx << 1) + 1;
			
			//NOT IMPLEMENTED
			//subb_L.ulcx = (ulcx + 1) >> 1;
			//subb_L.ulcy = (ulcy + 1) >> 1;
										
		}				
		else if (dimension == DIM_VER)
		{
			subb_L.orientation = WT_ORIENT_L;
			//subb_L.dimension = dim; UNSPEC!
			subb_L.level = level + 1;
			subb_L.ulx = ulx;
			subb_L.uly = uly;			
			subb_L.w = w;
			subb_L.h = (h + 1) >> 1;
			subb_L.gOrient = gOrient;
			subb_L.resLvl = (gOrient == WT_ORIENT_L) ? resLvl - 1 : resLvl;
			subb_L.anGainExp = anGainExp;
			subb_L.sbandIdx = (sbandIdx << 1); 
			
			
			subb_H.orientation = WT_ORIENT_H;
			//subb_H.dimension = dim; UNSPEC
			subb_H.gOrient = (gOrient == WT_ORIENT_L) ? WT_ORIENT_H : gOrient;
			subb_H.level = subb_L.level;
			subb_H.ulx = ulx;
			subb_H.uly = uly + subb_L.h;
			subb_L.w = w;
			subb_L.h = (h + 1) >> 1;
			subb_H.resLvl = resLvl;
			//gain 1 for highpass ==> lp + 0, lh + 1, hl + 1, hh + 2
			subb_H.anGainExp = anGainExp + 1; 		
			subb_H.sbandIdx = (sbandIdx << 1) + 1;
			
			//NOT IMPLEMENTED
			//subb_L.ulcx = (ulcx + 1) >> 1;
			//subb_L.ulcy = (ulcy + 1) >> 1;
										
		}
		else
		{
			throw new IllegalArgumentException("Only hor and ver dimensions implemented. If you want more, you have to implement it yourself :-) . ");
		}
		

	}

	/**
	 * Creates a Subband element with all the default values. The dimensions are
	 * (0,0), the upper left corner is (0,0) and the upper-left corner with
	 * respect to the canvas is (0,0) too.
	 */
	public SubbandAniso() {
	}
	
	
    /**
     * Creates a Subband Element with the given values, without doing any
     * dissection
     *
     * @param w The top-level width
     *
     * @param h The top-level height
     *
     * @param ulcx The horizontal coordinate of the upper-left corner with
     * respect to the canvas origin, in the component grid.
     *
     * @param ulcy The vertical coordinate of the upper-left corner with
     * respect to the canvas origin, in the component grid.
     *
     * @param lvls The number of levels (or LL decompositions) in the tree.
     * 
     * @see WaveletTransform
     * 
     * */
    public SubbandAniso(int w, int h, int ulcx, int ulcy, int lvls) {
		this.w = w;
		this.h = h;
		this.ulcx = ulcx;
		this.ulcy = ulcy;

		this.resLvl = lvls;
    }
    	   
    
    /**
     * Creates the top-level node and the entire subband tree, with the
     * top-level dimensions, the number of decompositions. 
     * [dyadic or full decomposition] 
     *
     * <P>For the analysis subband gain calculation it is assumed that
     * analysis filters are normalized with a DC gain of 1 and a Nyquist gain
     * of 2.
     *
     * <P>This constructor does not initialize the value of the magBits member
     * variable. This variable is normally initialized by the quantizer, on
     * the encoder side, or the bit stream reader, on the decoder side.
     *
     * @param w The top-level width
     *
     * @param h The top-level height
     *
     * @param ulcx The horizontal coordinate of the upper-left corner with
     * respect to the canvas origin, in the component grid.
     *
     * @param ulcy The vertical coordinate of the upper-left corner with
     * respect to the canvas origin, in the component grid.
     *
     * @param lvls The number of levels (or LL decompositions) in the tree.
     *
     * @param hfilters The horizontal wavelet filters (analysis or synthesis)
     * for each resolution level, starting at resolution level 0. If there are
     * less elements in the array than there are resolution levels, the last
     * element is used for the remaining resolution levels.
     *
     * @param vfilters The vertical wavelet filters (analysis or synthesis)
     * for each resolution level, starting at resolution level 0. If there are
     * less elements in the array than there are resolution levels, the last
     * element is used for the remaining resolution levels.
     * 
     * @param wt_type type of wavelet decomposition (dyadic of full)
     *
     * @see WaveletTransform
     * */
    public SubbandAniso(int w, int h, int ulcx, int ulcy, int lvls,
                   WaveletFilter hfilters[], WaveletFilter vfilters[], int wt_type) {

		// Initialize top-level node
		this.w = w;
		this.h = h;
		this.ulcx = ulcx;
		this.ulcy = ulcy;

		this.resLvl = lvls;
		
		switch (wt_type) {
		case WT_DYADIC:
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Dyadic Decomposition");
			break;
		case WT_FULL:
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Full Decomposition");
			break;
		default:
			break;
		}
		
		
		dissect(lvls, hfilters, vfilters, wt_type);

		plotTreeStructure("tree.eps", 512, 512);

    }
    
    /**
	 * create test tree structure down to level l
	 * 
	 * @param l
	 * @param structure
	 * @param hfilters
	 * @param vfilters
	 */
 	private void dissect (int l,  WaveletFilter hfilters[], WaveletFilter vfilters[], int wt_type)
	    {
	    	if (l == 0)
	    	{
	    		return;    		
	    	}
	    	
	    	int hi, vi;
	    	
	    	if (hfilters != null && vfilters != null)
	    	{
	    		//assign the approriate filter for this depth; if the array contains fewer
	    		//filters than the current resLvl, the last filter is applied
	    		hi = (resLvl <= hfilters.length) ? resLvl - 1 : hfilters.length-1;
	    		vi = (resLvl <= vfilters.length) ? resLvl - 1 : vfilters.length-1;
	    		split(DIM_HOR, hfilters[hi], vfilters[vi]);
	    	}
	    	else { //SubbandStructure

	    	if (orientation ==  WT_ORIENT_L) {
	    		if (l % 2 == 1)
	    		{
	    			split(DIM_VER, null, null);
	    		}
	    		else
	    		{
		    		split(DIM_HOR, null, null);
	    		}}
	    	else
	    	{
	    		if (l % 3 == 0)
	    		{
	    			split(DIM_VER, null, null);
	    		}
	    		else
	    		{
		    		split(DIM_HOR, null, null);
    		}}}
	    		    		    	    		    	
	    	getL().dissect (l-1, hfilters, vfilters, wt_type);
	    	getH().dissect (l-1, hfilters, vfilters, wt_type);
	    	
	    }

 	
	

	/**
	 * Creates the top-level node and the entire subband tree, with the
	 * top-level dimensions, the number of decompositions, and the decomposition
	 * tree as specified by wpStructure.
	 * 
	 * <P>
	 * For the analysis subband gain calculation it is assumed that analysis
	 * filters are normalized with a DC gain of 1 and a Nyquist gain of 2.
	 * 
	 * <P>
	 * This constructor does not initialize the value of the magBits member
	 * variable. This variable is normally initialized by the quantizer, on the
	 * encoder side, or the bit stream reader, on the decoder side.
	 * 
	 * @param w
	 *            The top-level width
	 * 
	 * @param h
	 *            The top-level height
	 * 
	 * @param ulcx
	 *            The horizontal coordinate of the upper-left corner with
	 *            respect to the canvas origin, in the component grid.
	 * 
	 * @param ulcy
	 *            The vertical coordinate of the upper-left corner with respect
	 *            to the canvas origin, in the component grid.
	 * 
	 * @param lvls
	 *            The number of levels (or LL decompositions) in the tree.
	 * 
	 * @param hfilters
	 *            The horizontal wavelet filters (analysis or synthesis) for
	 *            each resolution level, starting at resolution level 0. If
	 *            there are less elements in the array than there are resolution
	 *            levels, the last element is used for the remaining resolution
	 *            levels.
	 * 
	 * @param vfilters
	 *            The vertical wavelet filters (analysis or synthesis) for each
	 *            resolution level, starting at resolution level 0. If there are
	 *            less elements in the array than there are resolution levels,
	 *            the last element is used for the remaining resolution levels.
	 * 
	 * @param wpStructure
	 * 			  The wavelet packet decomposition structure as a bit pattern con-
	 * 		      tained in a byte buffer.
	 * 
	 * @see WaveletTransform
	 */
 	/*
	public SubbandAniso(int w, int h, int ulcx, int ulcy, int lvls,
			WaveletFilter hfilters[], WaveletFilter vfilters[], byte[] wpStructure) {
		
		// Initialize top-level node
		this.w = w;
		this.h = h;
		this.ulcx = ulcx;
		this.ulcy = ulcy;
		this.resLvl = lvls;
		int bitpos = 0;
		
		if (wpStructure == null)
		{
			throw new IllegalArgumentException ("no decomposition structure!!");
			/*FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "Dyadic Decomposition [passed wpStructure is empty]");
			dissect(lvls, hfilters, vfilters, WT_DYADIC);
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, ArrayUtil.printBitArrayAsHex(this.toBitPattern()));
			return;*/
/*

}
		
		bitpos = dissect_wp(lvls, hfilters, vfilters, wpStructure, 0);
		if (bitpos/8 != wpStructure.length -1)
		{
			FacilityManager.getMsgLogger().printmsg(MsgLogger.WARNING, "Did not use all of Wavelet Packest Specification String");
		}

		FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO, "WP Decomposition " + " [" + this.getClass().getName() + "]\n" + ArrayUtil.printBitArrayAsHex(wpStructure) );		
		plotTreeStructure("tree.eps", 512, 512);
	}
	*/
	
	
   /**
    * create tree structure down to level l
    * 
    * @param l current decomposition level 
    * @param hfilters Array of horizontal Filters to use
    * @param vfilters Array of vertical Filters to use
    * @param wpStructure wavelet packet decomposition structure as bitArray
    * @param bitpos current position in wpStructure
    * @return new position in wpStructure
    */
	
	/*
	private int dissect_wp(int l, WaveletFilter hfilters[],
			WaveletFilter vfilters[], byte[] wpStructure, int bitpos) {	
		
		// the global LL subband must not be decomposed further than the specified
		// resolution level. (This follows from the definition of resolution level) 		
		if (l < 0 && gOrient == WT_ORIENT_L)
		{
			throw new Error ("Could not create wavelet decomposition [global LL is below specified resLvl]. (Possibly something wrong with your wavelet decomposition specification?");
		}
		
		int bp = bitpos;
		if (ArrayUtil.readBit(wpStructure, bp++))
		{
	
			int hi, vi;
	
	
			if (hfilters != null && vfilters != null) {
				// assign the approriate filter for this depth; if the array
				// contains fewer
				// filters than the current resLvl, the last filter is applied
				hi = (resLvl <= hfilters.length) ? resLvl - 1 : hfilters.length - 1;
				vi = (resLvl <= vfilters.length) ? resLvl - 1 : vfilters.length - 1;
				split(hfilters[hi], vfilters[vi]);
			} else { //SubbandStructure
				split(null, null);
			}
	
			//split according to wpStructure						
			bp = getLL().dissect_wp(l - 1, hfilters, vfilters, wpStructure, bp);
			bp = getHL().dissect_wp(l - 1, hfilters, vfilters, wpStructure, bp);
			bp = getLH().dissect_wp(l - 1, hfilters, vfilters, wpStructure, bp);
			bp = getHH().dissect_wp(l - 1, hfilters, vfilters, wpStructure, bp);					
		}		
		return bp;
	}
	*/
	
	/**
	 * get rootsubband
	 * 
	 * @return
	 */
	public SubbandAniso getRoot() {
		SubbandAniso sb;
		sb = this;
		//go up to root
		while (sb.getParent() != null) {
			sb = sb.getParent();
		}
		return sb;
	}



	/**
	 * get the first lowest (i.e. smallest) leaf in a resolution level 
	 * 
	 * @param rl
	 * @return
	 */
	public SubbandAniso getLowestLeafInRes(int rl) {
		SubbandAniso sb;
		sb = getRoot();
		while (sb.resLvl != rl) //assumes that LL is at deepest resolution 
			//[which, by definition of resolution level, it is
		{
			sb = sb.getL();
		}
		//now we have the lowest ll subband in a resolution in sb

		//all that remains to do is to traverse to the HL daughter and then go
		// down to
		//it's lowest leaf (except if we are at the lowest resolution, which
		// consists
		//only of the lowest LL subband: in this case we return sb)
		if (rl == 0) {
			return sb;
		}

		return sb.getH().getLowestLeaf();

	}

	/**
	 * get the deepest LL-Subband in the _global_ LL tree (i.e. there might be 
	 * LL-subbands in other trees that are deeper!)
	 * 
	 * @return deepest (i.e. spatially smallest) subband with global orientation 
	 * of orientation WT_ORIENT_LL
	 */
	public SubbandAniso getDeepestGlobalLL() {

		SubbandAniso sb;
		sb = this;

		sb = getRoot();

		//go down LL
		while (sb.isNode) {
			sb = sb.getL();
		}
		return sb;
	}

	/**
	 * Return the next subband that is a child of this subband and a leaf
	 * 
	 * @return
	 */
	public SubbandAniso getLowestLeaf() {
		SubbandAniso sb;
		sb = this;

		while (sb.isNode) {
			sb = sb.getL();
		}
		return sb;
	}

	/**
	 * get minimum and maximum index of subbands in all resolution levels
	 * SIDNOTE maybe the black holes will have to be fixed index 0 = min index 1 =
	 * max index 2 = num
	 * 
	 * @return
	 */
	public int[][] findSubInResLvl() {
		int[][] subRange = new int[getRoot().resLvl + 1][3];

		SubbandAniso sb = getDeepestGlobalLL();
		SubbandAniso sb2;
		

		if (sb.resLvl != 0) {
			throw new Error(
					"Deepest global LL subband has to be at lowest resolution.");
			// by definition the deepest global LL subband has be at resolution level 0
			// if this is not the case, something is wrong.
		}

		subRange[0][0] = 0;
		subRange[0][1] = 0;
		subRange[0][2] = 1;

		//starting with 1, up to mrl, find minimum and maximum index for each
		// resolution level
		for (int l = 1; l <= getRoot().resLvl; l++) {
			sb = sb.getNextResLevel();
			sb2 = sb;
			int num = 1;
			int maxsbi = 0;

			while (sb2.nextSubband() != null) {
				sb2 = sb2.nextSubband();
				if (sb2.sbandIdx > maxsbi) maxsbi = sb2.sbandIdx; //This has to be checked, 
				//because the HH subband structure could be more shallow than the others, 
				//leading to lower numbers in higher subbands
				num++;
			}

			subRange[l][0] = sb.sbandIdx;
			subRange[l][1] = maxsbi;
			subRange[l][2] = num;
			FacilityManager.getMsgLogger().printmsg(
					MsgLogger.INFO,
					"Res.level " + l + " min idx = " + subRange[l][0]
							+ " max idx = " + subRange[l][1] + " num sbbs = "
							+ subRange[l][2]);
		}

		return subRange;

	}

	/**
	 * Returns the next subband in the same resolution level, following the
	 * subband index order. If already at the last subband then null is
	 * returned. If this subband is not a leaf an IllegalArgumentException is
	 * thrown.
	 * 
	 * 
	 * @return The next subband in the same resolution level, following the
	 *         subband index order, or null if already at last subband.
	 */
	public SubbandAniso nextSubband() {
		SubbandAniso sb;
				
		if (isNode) {
			throw new IllegalArgumentException();
		}

		switch (orientation) {
		case WT_ORIENT_L:
			sb = getParent();
			if (sb == null || sb.resLvl != resLvl) {
				// Already at top-level or last subband in res. level
				return null;
			} else {
				return sb.getH().getLowestLeaf();
			}
		case WT_ORIENT_H:
			// This is the complicated one
			sb = this;
			while (sb.orientation == WT_ORIENT_H) {
				sb = sb.getParent();
			}
			
			//now we in the L-Tree again			
			sb = sb.getParent();
			if (sb == null || sb.resLvl != resLvl) {
					// Already at top-level or last subband in res. level
					return null;
				} else {
					//go down this road
					sb = sb.getH().getLowestLeaf();
				}
			return sb;
		default:
			throw new Error("You have found a bug in JJ2000");
		}
	}

	/**
	 * return the tree-depth of a resolution level
	 * 
	 * @param r
	 * @return
	 */
	/*private int resMaxDepth(int rl) {

		Subband sb, sb2;

		// Find the root subband for the resolution level
		if (rl > resLvl || rl < 0) {
			throw new IllegalArgumentException();
		}
		sb = this;
		while (sb.resLvl > rl) {
			sb = sb.getLL();
		}

		return sb.resDepthRecursive();
	}*/

	
	/**
	 * get the maximum depth of decompsition levels
	 */
	private int maxDecompositionDepthRecursive() {
		if (!isNode) {
			return level;
		}

		return Math.max(getL().maxDecompositionDepthRecursive(), getH().maxDecompositionDepthRecursive());
	}
		
	/**
	 * Returns the first leaf subband element in the next higher resolution
	 * level.
	 * 
	 * seems to be WP proof
	 * 
	 * @return The first leaf element in the next higher resolution level, or
	 *         null if there is no higher resolution level.
	 *  
	 */
	public SubbandAniso getNextResLevel() {
		SubbandAniso sb;

		if (level == 0) { // No higher res. level
			return null;
		}
		// Go up until we get to a different resolution level
		sb = this;
		do {
			sb = sb.getParent();
			if (sb == null) { // No higher resolution level
				return null;
			}
		} while (sb.resLvl == resLvl);
		// Now go down to HL, which is in next higher resolution level
		sb = sb.getH();
		
		//and now get the first leaf
		sb = sb.getLowestLeaf();
		return sb;
	}

	/**
	 * Returns a subband element in the tree, given its resolution level and
	 * subband index. This method searches through the tree.
	 * 
	 * @param rl
	 *            The resolution level.
	 * 
	 * @param sbi
	 *            The subband index, within the resolution level.
	 *  
	 */
	public SubbandAniso getSubbandByIdx(int rl, int sbi) {
		SubbandAniso sb, sb2;
		int d, ci;

		if (rl > resLvl || rl < 0) {
			throw new IllegalArgumentException();
		}

		// Find the root subband for the resolution level
		sb = this;
		while (sb.resLvl > rl) {
			sb = sb.getL();
		}

		//SIDENOTE For now this is done recursively, just because it is simpler
		// to implement
		//in later stages, this should be changed to something more
		// sophisticated

		sb = sb.getSubbandByIdxRecursive(rl, sbi);
		

		if (sb.resLvl != rl || sb.sbandIdx != sbi) {
			throw new IllegalArgumentException("Error finding subband");
		}
		// Return subband
		return sb;
	}

	/**
	 * recursively look for a subband with index sbi in all subbands of a
	 * resolution
	 * 
	 * @param sbi
	 * @return
	 */
	private SubbandAniso getSubbandByIdxRecursive(int rl, int sbi) {

		SubbandAniso sb;
		if (!isNode) {
			if (sbandIdx == sbi) {
				return this;
			} else {
				return null;
			}
		}

		if (resLvl != rl)
			return null;

		sb = getL().getSubbandByIdxRecursive(rl, sbi);
		if (sb != null)
			return sb;

		sb = getH().getSubbandByIdxRecursive(rl, sbi);
		return sb;

	}
	
	/**
	 * recursively look for a subband with index sbi the whole subbandtree 
	 * 
	 * @param sbi
	 * @return
	 */
	public SubbandAniso findSbIdxInTree(int sbi) {

		SubbandAniso sb;
		if (!isNode) {
			if (sbandIdx == sbi) {
				return this;
			} else {
				return null;
			}
		}

		sb = getL().findSbIdxInTree(sbi);
		if (sb != null)
			return sb;
	
		sb = getH().findSbIdxInTree(sbi);
		return sb;

	}

	/**
	 * Returns a reference to the Subband element to which the specified point
	 * belongs. The specified point must be inside this (i.e. the one defined by
	 * this object) subband. This method searches through the tree.
	 * 
	 * @param x
	 *            horizontal coordinate of the specified point.
	 * 
	 * @param y
	 *            horizontal coordinate of the specified point.
	 */
	public SubbandAniso getSubband(int x, int y) {
		SubbandAniso cur, hhs;

		// Check that we are inside this subband
		if (x < ulx || y < uly || x >= ulx + w || y >= uly + h) {
			throw new IllegalArgumentException();
		}
/*
		cur = this;
		while (cur.isNode) {
			hhs = cur.getHH();
			// While we are still at a node -> continue
			if (x < hhs.ulx) {
				// Is the result of horizontal low-pass
				if (y < hhs.uly) {
					// Vertical low-pass
					cur = cur.getLL();
				} else {
					// Vertical high-pass
					cur = cur.getLH();
				}
			} else {
				// Is the result of horizontal high-pass
				if (y < hhs.uly) {
					// Vertical low-pass
					cur = cur.getHL();
				} else {
					// Vertical high-pass
					cur = cur.getHH();
				}
			}
		}

		return cur;
		*/
		throw new IllegalStateException("Implement me before using me :-P");

	}

	/**
	 * Returns subband informations in a string.
	 * 
	 * @return Subband informations
	 */
	public String toString() {

		String string = "[Subband #(id=" + sbandIdx + ", dim=" + dimension + ", rl=" + resLvl + "), " + "w=" + w
				+ ", h=" + h + ", ulx=" + ulx + ", uly=" + uly + ", ulcx= "
				+ ulcx + ", ulcy=" + ulcy + ", idx=" + sbandIdx + "\norient="
				+ orientation + ", node=" + isNode + ", level=" + level
				+ ", resLvl=" + resLvl + ", nomCBlkW=" + nomCBlkW
				+ ", nomCBlkH=" + nomCBlkH + "]";

		return string;
	}

	/**
	 * Returns subband information in a string, recursing through the whole
	 * subband tree
	 * 
	 * @return Subband information of the whole tree
	 */
	public String toStringRecurs(int indent) {

		String string = "";
		string += (isNode ? " |" : " <>") + " ";
		for (int i = 0; i < indent; i++) {
			string += "*";
		}
		string += " " + toString() + "\n";

		if (isNode) {
			string += "H " + getH().toStringRecurs(indent + 1);
			string += "L " + getL().toStringRecurs(indent + 1);		
		}
		return string;

	}

	/**
	 * Plot tree structure to eps file
	 * 
	 * (taken over from Ganesh++, originally taken from Wickerhauser 1994)
	 * 
	 * @param fname
	 *            output file
	 *  
	 */
	public void plotTreeStructure(String fname, int plotWidth, int plotHeight) {

		try {
			//create a buffered output stream
			FileOutputStream fos = new FileOutputStream(fname);
			BufferedOutputStream bout = new BufferedOutputStream(fos);
			DataOutputStream out = new DataOutputStream(bout);

			out.writeBytes("%!PS-Adobe-2.0 EPSF-2.0\n%%BoundingBox: 0 0");
			out.writeBytes(" " + plotWidth + " " + plotHeight);
			out.writeBytes("\n%%EndComments\ngsave\n");
			out.writeBytes("0.93 setgray\n");
			out.writeBytes("/Times-Roman findfont\n6 scalefont\nsetfont\n");
			out.writeBytes("0 setlinewidth\n");
			out.writeBytes("newpath 0 0 moveto " + plotWidth + " 0 lineto "
					+ plotWidth + " " + plotHeight + " lineto 0 " + plotHeight
					+ " lineto closepath fill stroke\n");
			plotTreeStructureRecursive(out, plotWidth, plotHeight, plotWidth,
					0, 0, 0.86f, true);
			out.writeBytes("stroke\ngrestore\n");
			out.flush();
			out.close();
		} catch (FileNotFoundException fnfe) {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.ERROR,
					"File not Found");
		} catch (IOException ioe) {
			FacilityManager.getMsgLogger()
					.printmsg(MsgLogger.ERROR, "IO Error");
		} finally {
			FacilityManager.getMsgLogger().printmsg(MsgLogger.INFO,
					"Tree structure written to " + fname);
		}
	}

	/**
	 * recursively plot subband structure
	 * 
	 * @param out
	 * @param width
	 * @param height
	 * @param plotWidth
	 * @param plotHeight
	 * @param x
	 * @param y
	 * @param gray
	 */
	private void plotTreeStructureRecursive(DataOutputStream out, int width,
			int height, int plotWidth, int x, int y, float gray,
			boolean showNumbers) {

		try {
			int w2, h2;
			int xh, yh;
			
			
			//reversed dimensions in gnuplot!
			if (dimension == DIM_VER)
			{
				w2 = width / 2;
				h2 = height;
				xh = x+w2;
				yh = y;
			}
			else
			{
				w2 = width;
				h2 = height / 2;		
				xh = x;
				yh = y+h2;
			}

			if (!isNode)

			{
				if (showNumbers) {
					//out.writeBytes("grestore "
					out.writeBytes("0 setgray\n");
					out.writeBytes("newpath\n" + (y + 2) + " "
							+ (plotWidth - (x + 6)) + " moveto\n");
					out.writeBytes("(" + sbandIdx + "/" + resLvl + ") show\n");
					/*out.writeBytes("" + (y + 2) + " " + (plotWidth - (x + 13))
							+ " moveto\n");
					out.writeBytes("(" + resLvl + "/" + level + ") show\n");*/

				}
				return;
			} else {

				out.writeBytes("% plotting " + this.sbandIdx + "\n");
				out.writeBytes(Float.toString(gray) + " setgray\n");
				out.writeBytes("newpath " + y + " " + (plotWidth - x)
						+ " moveto " + y + " " + (plotWidth - (x + width))
						+ " lineto");
				out.writeBytes(" " + (y + height) + " "
						+ (plotWidth - (x + width)) + " lineto " + (y + height)
						+ " " + (plotWidth - x)
						+ " lineto closepath gsave fill stroke\n");

				out.writeBytes("grestore 0 setgray\n");
				out.writeBytes("" + (y + h2) + " " + (plotWidth - x)
						+ " moveto " + (y + h2) + " "
						+ (plotWidth - (x + width)) + " lineto\n");
				out.writeBytes("" + y + " " + (plotWidth - (x + w2))
						+ " moveto " + (y + height) + " "
						+ (plotWidth - (x + w2)) + " lineto stroke\n");

				if (gOrient == WT_ORIENT_L) gray -= 0.07;

				getL().plotTreeStructureRecursive(out, w2, h2, plotWidth, x,
						y, gray, showNumbers);
				getH().plotTreeStructureRecursive(out, w2, h2, plotWidth, xh,
						yh, gray, showNumbers);
			}
		} catch (IOException ioe) {
			FacilityManager.getMsgLogger()
					.printmsg(MsgLogger.ERROR, "IO Error");
		}
	}

	/**
	 * This function returns the horizontal wavelet filter relevant to this
	 * subband
	 * 
	 * @return The horizontal wavelet filter
	 */
	public abstract WaveletFilter getHorWFilter();

	/**
	 * This function returns the vertical wavelet filter relevant to this
	 * subband
	 * 
	 * @return The vertical wavelet filter
	 */
	public abstract WaveletFilter getVerWFilter();

}

