/*
 * CVS identifier:
 *
 * $Id: SynWTFilterFloatHaarSym.java,v 1.2 2006-03-28 08:01:21 dengel Exp $
 *
 * Class:                   SynWTFilterFloatHaar
 *
 * Description:             A synthetizing wavelet filter implementing the
 *                          lifting Haar transform.
 *
 *
 *
 * COPYRIGHT:
 * 
 * This software module was originally developed by Raphal Grosbois and
 * Diego Santa Cruz (Swiss Federal Institute of Technology-EPFL); Joel
 * Askelf (Ericsson Radio Systems AB); and Bertrand Berthelot, David
 * Bouchard, Flix Henry, Gerard Mozelle and Patrice Onno (Canon Research
 * Centre France S.A) in the course of development of the JPEG2000
 * standard as specified by ISO/IEC 15444 (JPEG 2000 Standard). This
 * software module is an implementation of a part of the JPEG 2000
 * Standard. Swiss Federal Institute of Technology-EPFL, Ericsson Radio
 * Systems AB and Canon Research Centre France S.A (collectively JJ2000
 * Partners) agree not to assert against ISO/IEC and users of the JPEG
 * 2000 Standard (Users) any of their rights under the copyright, not
 * including other intellectual property rights, for this software module
 * with respect to the usage by ISO/IEC and Users of this software module
 * or modifications thereof for use in hardware or software products
 * claiming conformance to the JPEG 2000 Standard. Those intending to use
 * this software module in hardware or software products are advised that
 * their use may infringe existing patents. The original developers of
 * this software module, JJ2000 Partners and ISO/IEC assume no liability
 * for use of this software module or modifications thereof. No license
 * or right to this software module is granted for non JPEG 2000 Standard
 * conforming products. JJ2000 Partners have full right to use this
 * software module for his/her own purpose, assign or donate this
 * software module to any third party and to inhibit third parties from
 * using this software module for non JPEG 2000 Standard conforming
 * products. This copyright notice must be included in all copies or
 * derivative works of this software module.
 * 
 * Copyright (c) 1999/2000 JJ2000 Partners.
 *  */
package jj2000.j2k.wavelet.synthesis;

import jj2000.j2k.wavelet.*;


/**
 * This class inherits from the synthesis wavelet filter definition for int
 * data. It implements the inverse wavelet transform specifically for the 9x7
 * filter. The implementation is based on the lifting scheme.
 *
 * <P>See the SynWTFilter class for details such as normalization, how to
 * split odd-length signals, etc. In particular, this method assumes that the
 * low-pass coefficient is computed first.
 *
 * @see SynWTFilter
 * @see SynWTFilterFloat
 * */
public class SynWTFilterFloatHaarSym extends SynWTFilterFloatHaar {
    
    /**
     * An implementation of the synthetize_lpf() method that works on int
     * data, for the inverse 9x7 wavelet transform using the lifting
     * scheme. See the general description of the synthetize_lpf() method in
     * the SynWTFilter class for more details.
     *
     * <P>The low-pass and high-pass subbands are normalized by respectively a
     * factor of 1/KL and a factor of 1/KH
     *
     *
     * @param lowSig This is the array that contains the low-pass input
     * signal.
     *
     * @param lowOff This is the index in lowSig of the first sample to
     * filter.
     *
     * @param lowLen This is the number of samples in the low-pass input
     * signal to filter.
     *
     * @param lowStep This is the step, or interleave factor, of the low-pass
     * input signal samples in the lowSig array.
     *
     * @param highSig This is the array that contains the high-pass input
     * signal.
     *
     * @param highOff This is the index in highSig of the first sample to
     * filter.
     *
     * @param highLen This is the number of samples in the high-pass input
     * signal to filter.
     *
     * @param highStep This is the step, or interleave factor, of the
     * high-pass input signal samples in the highSig array.
     *
     * @param outSig This is the array where the output signal is placed. It
     * should be long enough to contain the output signal.
     *
     * @param outOff This is the index in outSig of the element where to put
     * the first output sample.
     *
     * @param outStep This is the step, or interleave factor, of the output
     * samples in the outSig array.
     *
     * @see SynWTFilter#synthetize_lpf
     * */
    public
        void synthetize_lpf(float[] lowSig, int lowOff, int lowLen, int lowStep,
                        float[] highSig, int highOff, int highLen, int highStep,
                        float[] outSig, int outOff, int outStep) {
                        
        int i;
        int outLen = lowLen + highLen; //Length of the output signal
        int iStep = 2*outStep; //Upsampling in outSig
        int ik; //Indexing outSig
        int lk; //Indexing lowSig
        int hk; //Indexing highSig
        
        SymbolicMatrix sm = SymbolicMatrix.getInstance();
        String[] lowSigS =  sm.getBuffer();
        String[] highSigS = sm.getBuffer();
        String[] outSigS = sm.data;

        
        // Generate even samples       

        //Initialize counters
        lk = lowOff;
        hk = highOff;
        ik = outOff;
        
        //Handle tail boundary effect. Use symmetric extension
        /*if(outLen>1) {
            outSig[ik] = lowSig[lk]/KL - 2*DELTA*highSig[hk]/KH;
        }
        else {
        	outSig[ik] = lowSig[lk];
        }
        
        lk += lowStep;
        hk += highStep;
    	ik += iStep;*/
    	
        //Apply lifting step to each "inner" sample
        for(i = 0; i<outLen-1; i+=2, ik+=iStep, lk+=lowStep, hk+=highStep) {
            outSig[ik] = lowSig[lk]/KL - 0.5f*(highSig[hk]/KH);
            
            //SYM
            outSigS[ik] = "(" + lowSigS[lk] + "- 0.5 " + highSigS[hk] + ")";
        }
        
        //Handle head boundary effect if input signal has odd length
        /*if(outLen%2 == 1) {
            if(outLen>2){
                outSig[ik] = lowSig[lk]/KL - 
                2*DELTA*highSig[hk-highStep]/KH;
            }
        }*/
        
        // Generate odd samples
         
        //Initialize counters
        lk = lowOff;
        hk = highOff;
    	ik = outOff + outStep;

        //Apply lifting step to each "inner" sample
        //for(i = 1; i<outLen-1; i+=2, ik+=iStep, hk+=highStep, lk+=lowStep) {
        for(i = 0; i<outLen-1; i+=2, ik+=iStep, lk+=lowStep, hk+=highStep) {        	
            outSig[ik] = outSig[ik-outStep] + highSig[hk]/KH;
            
            //SYM
            outSigS[ik] = "(" + outSigS[ik-outStep] + " + " +  highSigS[hk] + ")";
    	}

        //Handle head boundary effect if output signal has even length
        /*if(outLen % 2 == 0) {
            outSig[ik] = highSig[hk]/KH - 2*GAMMA*outSig[ik-outStep];
        } */      

    }
    
    /** 
     * Returns a string of information about the synthesis wavelet filter
     *
     * @return wavelet filter type.
     *
     *
     */
    public String toString(){
        return "HaarS";
    }
    
}
